#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Computation of modal coupling coefficients via the Bayer-Helms
analytic formalism."""

from libc.stdlib cimport calloc, free
from libc.string cimport memset

from cython.parallel import prange

cimport numpy as np
import numpy as np

from finesse.cymath.complex cimport carg, conj, creal, cimag, csqrt # Standard complex.h functions
from finesse.cymath.complex cimport ceq, inverse_unsafe, crotate, crotate2, cpow_re, cnorm
from finesse.cymath.math cimport exp, sqrt, cos, sin, ceil, floor # Standard math.h functions
from finesse.cymath.math cimport factorial, sqrt_factorial, msign, nmin, nmax
from finesse.cymath.gaussbeam cimport gouy, waist_size

from finesse.utilities.cyomp cimport determine_nthreads_even


from finesse.knm.matrix cimport KnmMatrix


### Python functions for computing Bayer-Helms coefficients and matrices ###
###            not associated with any component or workspace            ###

def compute_bayerhelms_coeff(
    qx1, qx2, qy1, qy2,
    double xgamma, double ygamma,
    int n1, int m1, int n2, int m2,
    double nr=1.0, double wavelength=1064e-9,
    bint reverse_gouy=False,
):
    """Computes the coupling coefficient, via Bayer-Helms :cite:`BayerHelms`,
    from mode (n1, m1) -> mode (n2, m2).

    .. note::

        Use :func:`.make_bayerhelms_matrix` to compute a scattering matrix of these
        coefficients rather than calling this function in a loop.

    Parameters
    ----------
    qx1 : :class:`.BeamParam` or complex
        Input beam parameter in the tangential plane.

    qx2 : :class:`.BeamParam` or complex
        Output beam parameter in the tangential plane.

    qy1 : :class:`.BeamParam` or complex
        Input beam parameter in the sagittal plane.

    qy2 : :class:`.BeamParam` or complex
        Output beam parameter in the sagittal plane.

    xgamma : float
        Misalignment angle in the tangential plane (in radians).

    ygamma : float
        Misalignment angle in the sagittal plane (in radians).

    n1 : int
        Input tangential mode index.

    m1 : int
        Input sagittal mode index.

    n2 : int
        Output tangential mode index.

    m2 : int
        Output sagittal mode index.

    nr : float, optional; default: 1.0
        Refractive index of the associated medium.

    wavelength : float, optional; default: 1064 nm
        Wavelength of the beam (in metres).

    Returns
    -------
    coeff : complex
        The complex coupling coefficient for the specified mode coupling.
    """
    cdef:
        knm_bh_workspace kws_x
        knm_bh_workspace kws_y
        int maxtem = nmax(n1 + m1, n2 + m2)

    knm_bh_ws_init(&kws_x, qx1, qx2, xgamma, 1.0, nr, wavelength, maxtem)
    knm_bh_ws_init(&kws_y, qy1, qy2, ygamma, 1.0, nr, wavelength, maxtem)

    cdef complex_t coeff = k_nmnm(&kws_x, &kws_y, n1, m1, n2, m2)
    if reverse_gouy:
        coeff = rev_gouy(&kws_x, &kws_y, coeff, n1, m1, n2, m2)

    knm_bh_ws_free(&kws_x)
    knm_bh_ws_free(&kws_y)

    return coeff


def make_bayerhelms_matrix(
    qx1, qx2, qy1, qy2,
    double xgamma, double ygamma,
    double nr=1.0, double wavelength=1064e-9,
    bint reverse_gouy=False,
    bint parallel=True,
    **mode_selection_kwargs,
):
    """Constructs and computes a coupling coefficient scattering
    matrix using the Bayer-Helms :cite:`BayerHelms` analytic
    formalism.

    See the function :func:`.make_modes` for the arguments which
    should be passed via `mode_selection_kwargs`.

    Parameters
    ----------
    qx1 : :class:`.BeamParam` or complex
        Input beam parameter in the tangential plane.

    qx2 : :class:`.BeamParam` or complex
        Output beam parameter in the tangential plane.

    qy1 : :class:`.BeamParam` or complex
        Input beam parameter in the sagittal plane.

    qy2 : :class:`.BeamParam` or complex
        Output beam parameter in the sagittal plane.

    xgamma : float
        Misalignment angle in the tangential plane (in radians).

    ygamma : float
        Misalignment angle in the sagittal plane (in radians).

    nr : float, optional; default: 1.0
        Refractive index of the associated medium.

    wavelength : float, optional; default: 1064 nm
        Wavelength of the beam (in metres).

    parallel : bool, optional; default: True
        Whether to enable multi-threaded execution via OpenMP. The
        number of threads used scales with the number of modes;
        bounded by the maximum thread count as given by
        ``omp_get_max_threads``.

    mode_selection_kwargs : keyword arguments
        See :func:`.make_modes`.

    Returns
    -------
    kmat : :class:`.KnmMatrix`
        The resulting scattering matrix as a :class:`.KnmMatrix` object.

    Examples
    --------
    See :ref:`arbitrary_scatter_matrices`.
    """
    from finesse.utilities.homs import make_modes

    cdef:
        knm_bh_workspace kws_x
        knm_bh_workspace kws_y

    modes = make_modes(**mode_selection_kwargs)
    cdef int[:, ::1] modes_view = modes
    cdef int maxtem = np.max(modes[:, 0] + modes[:, 1])

    cdef KnmMatrix kmat = KnmMatrix(modes_view)

    knm_bh_ws_init(&kws_x, qx1, qx2, xgamma, 1.0, nr, wavelength, maxtem)
    knm_bh_ws_init(&kws_y, qy1, qy2, ygamma, 1.0, nr, wavelength, maxtem)

    compute_knm_matrix_bh(&kws_x, &kws_y, modes_view, kmat.data_view, parallel)
    if reverse_gouy:
        reverse_gouy_phases(&kws_x, &kws_y, kmat.data_view, modes_view, kmat.data_view)

    knm_bh_ws_free(&kws_x)
    knm_bh_ws_free(&kws_y)

    return kmat


### Workspace for Bayer-Helms calculations ###

cdef void knm_bh_ws_init(
    knm_bh_workspace* kws,
    complex_t q1,
    complex_t q2,
    double beta,
    double beta_factor,
    double nr,
    double lambda0,
    int maxtem,
    bint is_mismatch_changing=True,
    bint is_alignment_changing=False
) nogil:
    kws.nr = nr
    kws.lambda0 = lambda0
    kws.beta_factor = beta_factor
    kws.is_mm_changing = is_mismatch_changing
    kws.is_alignment_changing = is_alignment_changing

    kws.pow_cache_size = maxtem + 1
    kws.F_pow_cache = <complex_t*> calloc(kws.pow_cache_size, sizeof(complex_t))
    kws.FS_pow_cache = <complex_t*> calloc(kws.pow_cache_size, sizeof(complex_t))
    kws.X_pow_cache = <complex_t*> calloc(kws.pow_cache_size, sizeof(complex_t))
    kws.XS_pow_cache = <complex_t*> calloc(kws.pow_cache_size, sizeof(complex_t))
    kws.K0_plus1_pow_cache = <double*> calloc(kws.pow_cache_size, sizeof(double))

    # Compute all of the workspace parameters - i.e. both mismatch
    # and misalignment terms, as well as other quantities like the
    # Gouy phases of q1, q2 required for reverse Gouy operations
    knm_bh_ws_recompute(kws, q1, q2, beta)

cdef void knm_bh_ws_free(knm_bh_workspace* kws) nogil:
    if kws.F_pow_cache != NULL: free(kws.F_pow_cache)
    if kws.FS_pow_cache != NULL: free(kws.FS_pow_cache)
    if kws.X_pow_cache != NULL: free(kws.X_pow_cache)
    if kws.XS_pow_cache != NULL: free(kws.XS_pow_cache)
    if kws.K0_plus1_pow_cache != NULL: free(kws.K0_plus1_pow_cache)

cdef void knm_bh_ws_reset_mm_cpow_caches(knm_bh_workspace* kws) nogil:
    cdef Py_ssize_t i

    memset(kws.F_pow_cache, 0, kws.pow_cache_size * sizeof(complex_t))
    memset(kws.FS_pow_cache, 0, kws.pow_cache_size * sizeof(complex_t))
    memset(kws.K0_plus1_pow_cache, 0, kws.pow_cache_size * sizeof(double))
    kws.F_pow_cache[0] = 1.0 + 0.0j
    kws.FS_pow_cache[0] = 1.0 + 0.0j

    cdef double root_K0_plus1 = sqrt(1.0 + kws.K0)
    kws.K0_plus1_pow_cache[0] = sqrt(root_K0_plus1)

    if not kws.matched:
        for i in range(1, kws.pow_cache_size):
            kws.F_pow_cache[i] = kws.F * kws.F_pow_cache[i - 1]
            kws.FS_pow_cache[i] = kws.FS * kws.FS_pow_cache[i - 1]
            kws.K0_plus1_pow_cache[i] = root_K0_plus1 * kws.K0_plus1_pow_cache[i - 1]

cdef void knm_bh_ws_reset_alignment_cpow_caches(knm_bh_workspace* kws) nogil:
    cdef Py_ssize_t i

    memset(kws.X_pow_cache, 0, kws.pow_cache_size * sizeof(complex_t))
    memset(kws.XS_pow_cache, 0, kws.pow_cache_size * sizeof(complex_t))
    kws.X_pow_cache[0] = 1.0 + 0.0j
    kws.XS_pow_cache[0] = 1.0 + 0.0j

    if not kws.aligned:
        for i in range(1, kws.pow_cache_size):
            kws.X_pow_cache[i] = kws.X * kws.X_pow_cache[i - 1]
            kws.XS_pow_cache[i] = kws.XS * kws.XS_pow_cache[i - 1]

cdef inline bint knm_bh_ws_is_changing(const knm_bh_workspace* kws) nogil:
    return kws.is_mm_changing or kws.is_alignment_changing

cdef void knm_bh_ws_recompute(knm_bh_workspace* kws, complex_t q1, complex_t q2, double beta) nogil:
    knm_bh_ws_recompute_mismatch(kws, q1, q2)
    knm_bh_ws_recompute_misalignment(kws, beta)

cdef void knm_bh_ws_recompute_mismatch(knm_bh_workspace* kws, complex_t q1, complex_t q2) nogil:
    cdef:
        double zr
        complex_t K

    kws.q1 = q1
    kws.q2 = q2

    kws.gouy1 = gouy(q1)
    kws.gouy2 = gouy(q2)

    kws.matched = ceq(q1, q2)
    if kws.matched:
        kws.K_conj = COMPLEX_0
        kws.k1 = 1.0 + 0.0j
        kws.K0 = 0.0
        kws.FS = COMPLEX_0
        kws.F = COMPLEX_0
    else:
        zr = 1.0 / cimag(q2)

        # K is the mismatch parameter
        K = 0.5 * zr * ((cimag(q1) - cimag(q2)) + (creal(q1) - creal(q2)) * 1.0j)
        kws.K_conj = conj(K)
        # The X, XS variables contain a 1/sqrt(1+K*) factor, this
        # is computed here then applied to kws.X, kws.XS later
        kws.k1 = inverse_unsafe(csqrt(1.0 + kws.K_conj))
        kws.K0 = 2.0 * creal(K)

        kws.F = 0.5 * kws.K_conj
        kws.FS = 0.5 * K / (1.0 + kws.K0)

    knm_bh_ws_reset_mm_cpow_caches(kws)

cdef void knm_bh_ws_recompute_misalignment(knm_bh_workspace* kws, double beta) nogil:
    cdef:
        double Theta, gg0, z2z0
        complex_t x1, x2, exponent

    kws.gamma = kws.beta_factor * beta
    kws.aligned = kws.gamma == 0.0
    if kws.aligned:
        kws.X = COMPLEX_0
        kws.XS = COMPLEX_0

        # These won't be used when system is aligned but here they
        # will be initialised to correct corresponding values in
        # case they get accessed accidentally at some point
        kws.ex = 1.0
        kws.ex_cos_phase = 1.0
        kws.ex_sin_phase = 0.0
    else:
        # TODO (sjr) The following should also include lateral displacement
        #            terms from Bayer-Helms (in X, XS, ex variables), see
        #            Issue #238 which tracks this

        # The divergence angle of the second mode
        Theta = waist_size(kws.q2, kws.nr, kws.lambda0) / cimag(kws.q2)
        # this gives the factor (gamma / w0) * zr
        gg0 = kws.gamma / Theta
        z2z0 = creal(kws.q2) / cimag(kws.q2)

        x1 = z2z0 - 1.0j
        x2 = z2z0 + (1.0 + 2.0 * kws.K_conj) * 1.0j

        # Here we have X  = (gamma / w0) * zr * (z2 / zr +i*(1 + 2K*)) * 1/sqrt(1 + K*)
        #                 = (gamma / w0) * (z2 + i*zr*(1 + 2K*)) / sqrt(1 + K*)
        kws.X = gg0 * x2 * kws.k1
        # And Xbar = (gamma / w0) * zr * (z2 / zr - i) * 1/sqrt(1 + K*)
        #          = (z2 - i*zr) * (gamma / w0) / sqrt(1 + K*)
        kws.XS = gg0 * x1 * kws.k1
        # This is the exponent of the E^(x) equation seen in Bayer-Helms...
        exponent = 0.5 * kws.XS * kws.X

        # ... and these correspond to the terms of E^(x)
        # i.e. exp(-ex) = exp(-Re(ex)) * exp(-i*Im(ex))
        #               = exp(-Re(ex)) * (cos(Im(ex)) - i*sin(Im(ex)))
        # we pre-calculate these for efficiency so that crotate2 can be used in k_mm
        kws.ex = exp(-creal(exponent))
        kws.ex_cos_phase = cos(cimag(exponent))
        kws.ex_sin_phase = -sin(cimag(exponent))

    knm_bh_ws_reset_alignment_cpow_caches(kws)


### Below are the functions to be used in the compute_scattering_matrices ###
###  function of KnmConnectorWorkspace, each is optimised to the extreme  ###


cdef void fast_compute_refl_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
):
    """Compute a reflection Bayer-Helms scattering matrix in a very highly
    optimised way. Does a single pass over the matrix: computing
    the coupling coefficients, reverse Gouy phases in-place and flips sign
    of all odd couplings in tangential plane.

    Uses a flat pointer to output matrix rather than memory-view such
    that KnmConnectorWorkspace.compute_scattering_matrices doesn't need
    to do any slow converting to memory-views from matrix pointers.
    """
    cdef:
        Py_ssize_t i, j

        int n1, m1, n2, m2
        complex_t coeff, coeff_rg, coeff_y_refl

    for i in prange(Nhoms, nogil=True, num_threads=nthreads, schedule="static"):
        n1 = homs[2 * i]
        m1 = homs[2 * i + 1]
        for j in range(Nhoms):
            n2 = homs[2 * j]
            m2 = homs[2 * j + 1]

            # Compute the coefficent per BH
            coeff = k_nmnm(kws_x, kws_y, n1, m1, n2, m2)
            # Reverse the Gouy phase in-place
            coeff_rg = rev_gouy(kws_x, kws_y, coeff, n1, m1, n2, m2)
            # Flip sign of odd-order tangential modes
            coeff_y_refl = msign(n2) * coeff_rg

            out[j*Nhoms + i] = coeff_y_refl

cdef void fast_compute_trns_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
):
    """Compute a transmission Bayer-Helms scattering matrix in a very highly
    optimised way. Does a single pass over the matrix: computing
    the coupling coefficients and reverse Gouy phases in-place.

    Uses a flat pointer to output matrix rather than memory-view such
    that KnmConnectorWorkspace.compute_scattering_matrices doesn't need
    to do any slow converting to memory-views from matrix pointers.
    """
    cdef:
        Py_ssize_t i, j

        int n1, m1, n2, m2
        complex_t coeff, coeff_rg

    for i in prange(Nhoms, nogil=True, num_threads=nthreads, schedule="static"):
        n1 = homs[2 * i]
        m1 = homs[2 * i + 1]
        for j in range(Nhoms):
            n2 = homs[2 * j]
            m2 = homs[2 * j + 1]

            # Compute the coefficent per BH
            coeff = k_nmnm(kws_x, kws_y, n1, m1, n2, m2)
            # Reverse the Gouy phase in-place
            coeff_rg = rev_gouy(kws_x, kws_y, coeff, n1, m1, n2, m2)

            out[j*Nhoms + i] = coeff_rg

cdef void fast_compute_refl_knm_matrix_bh__k00_zeroed(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
):
    """Compute a reflection Bayer-Helms scattering matrix in a very highly
    optimised way. Does a single pass over the matrix: computing
    the coupling coefficients, reverse Gouy phases in-place, zeroes
    the phase of the k0000 coeff and flips sign of all odd couplings
    in tangential plane.

    Uses a flat pointer to output matrix rather than memory-view such
    that KnmConnectorWorkspace.compute_scattering_matrices doesn't need
    to do any slow converting to memory-views from matrix pointers.
    """
    cdef:
        Py_ssize_t i, j

        int n1, m1, n2, m2
        complex_t coeff, coeff_rg, coeff_rotate, coeff_y_refl

        complex_t k00 = k_nmnm(kws_x, kws_y, 0, 0, 0, 0)
        double phase_k00 = carg(rev_gouy(kws_x, kws_y, k00, 0, 0, 0, 0))
        double cph = cos(phase_k00)
        double sph = -sin(phase_k00)

    for i in prange(Nhoms, nogil=True, num_threads=nthreads, schedule="static"):
        n1 = homs[2 * i]
        m1 = homs[2 * i + 1]
        for j in range(Nhoms):
            n2 = homs[2 * j]
            m2 = homs[2 * j + 1]

            # Compute the coefficent per BH
            coeff = k_nmnm(kws_x, kws_y, n1, m1, n2, m2)
            # Reverse the Gouy phase in-place
            coeff_rg = rev_gouy(kws_x, kws_y, coeff, n1, m1, n2, m2)
            # Rotate the coefficient by the negative phase of k0000
            coeff_rotate = crotate2(coeff_rg, cph, sph)
            # Flip sign of odd-order tangential modes
            coeff_y_refl = msign(n2) * coeff_rotate

            out[j*Nhoms + i] = coeff_y_refl

cdef void fast_compute_trns_knm_matrix_bh__k00_zeroed(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
):
    """Compute a transmission Bayer-Helms scattering matrix in a very highly
    optimised way. Does a single pass over the matrix: computing
    the coupling coefficients, reverse Gouy phases in-place and zeroes
    the phase of the k0000 coeff.

    Uses a flat pointer to output matrix rather than memory-view such
    that KnmConnectorWorkspace.compute_scattering_matrices doesn't need
    to do any slow converting to memory-views from matrix pointers.
    """
    cdef:
        Py_ssize_t i, j

        int n1, m1, n2, m2
        complex_t coeff, coeff_rg, coeff_rotate

        complex_t k00 = k_nmnm(kws_x, kws_y, 0, 0, 0, 0)
        double phase_k00 = carg(rev_gouy(kws_x, kws_y, k00, 0, 0, 0, 0))
        double cph = cos(phase_k00)
        double sph = -sin(phase_k00)

    for i in prange(Nhoms, nogil=True, num_threads=nthreads, schedule="static"):
        n1 = homs[2 * i]
        m1 = homs[2 * i + 1]
        for j in range(Nhoms):
            n2 = homs[2 * j]
            m2 = homs[2 * j + 1]

            # Compute the coefficent per BH
            coeff = k_nmnm(kws_x, kws_y, n1, m1, n2, m2)
            # Reverse the Gouy phase in-place
            coeff_rg = rev_gouy(kws_x, kws_y, coeff, n1, m1, n2, m2)
            # Rotate the coefficient by the negative phase of k0000
            coeff_rotate = crotate2(coeff_rg, cph, sph)

            out[j*Nhoms + i] = coeff_rotate


### Function for computing BH matrix with no modifications. ###
### The modification functions are defined separately here. ###


cdef void compute_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int[:, ::1] homs,
    complex_t[:, ::1] out,
    bint parallel=True,
):
    """Compute a Bayer-Helms scattering matrix.

    This is used only for the Python utility function `make_bayerhelms_matrix`. The
    heavily optimised versions, which get used by KnmConnectorWorkspaces during
    simulations, are above.
    """
    cdef:
        Py_ssize_t N = homs.shape[0]
        Py_ssize_t i, j

        int n1, m1, n2, m2

        int nthreads = 0

    if parallel:
        # See CarrierSignalMatrixSimulation.initialise_sim_config_data
        # for details on the thread count setting logic used here
        nthreads = determine_nthreads_even(N, 10)

    if not nthreads:
        nthreads = 1

    for i in prange(N, nogil=True, num_threads=nthreads, schedule="static"):
        n1 = homs[i][0]
        m1 = homs[i][1]
        for j in range(N):
            n2 = homs[j][0]
            m2 = homs[j][1]
            out[j][i] = k_nmnm(kws_x, kws_y, n1, m1, n2, m2)


cdef void zero_tem00_phase(
    const complex_t[:, ::1] knm_mat,
    complex_t[:, ::1] out
) nogil:
    """Rotates all coupling coefficients in the matrix `knm_mat`
    by the phase of the zeroth (00 -> 00) coupling coefficient.

    This has the effect of zeroing the phase of the 00 -> 00 coupling
    coefficient, resulting in cavities being resonant for the 00 mode
    by default if this function (along with zeroing of space Gouy phase
    for 00 mode) is applied."""
    cdef:
        Py_ssize_t i, j
        Py_ssize_t N = knm_mat.shape[0]
        double phase_k0000 = carg(knm_mat[0][0])
        double cph = cos(phase_k0000)
        double sph = -sin(phase_k0000)

    for i in range(N):
        for j in range(N):
            out[i][j] = crotate2(knm_mat[i][j], cph, sph)


cdef void flip_odd_horizontal(
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) nogil:
    """Flips the sign of all odd couplings in the sagittal plane."""
    cdef:
        Py_ssize_t i, j
        int n2
        Py_ssize_t N = knm_mat.shape[0]

    for i in range(N):
        for j in range(N):
            n2 = homs[j][0]

            if msign(n2) == -1:
                out[j][i] = -knm_mat[j][i]


cdef void reverse_gouy_phases(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) nogil:
    """
    Adjust the phase of all coupling coefficients in the matrix `knm_mat` with
    respect to the Gouy phases.

    This is required for :math:`k_{nmn'm'}` calculations because in Finesse the Gouy phase
    is added explicitly to the amplitude coefficients in a :class:`.Space` whereas
    the coupling coefficients are derived using a formula in which the Gouy phase
    resides in the equation for the spatial profile.
    """
    cdef:
        Py_ssize_t i, j
        int n1, m1, n2, m2
        Py_ssize_t N = knm_mat.shape[0]

    for i in range(N):
        n1 = homs[i][0]
        m1 = homs[i][1]
        for j in range(N):
            n2 = homs[j][0]
            m2 = homs[j][1]

            out[j][i] = rev_gouy(
                kws_x, kws_y,
                knm_mat[j][i],
                n1, m1, n2, m2,
            )

cdef complex_t rev_gouy(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    complex_t k,
    int n1, int m1, int n2, int m2,
) nogil:
    cdef:
        double phase1, phase2

    phase1 = (n1 + 0.5) * kws_x.gouy1 + (m1 + 0.5) * kws_y.gouy1
    phase2 = (n2 + 0.5) * kws_x.gouy2 + (m2 + 0.5) * kws_y.gouy2

    return crotate(k, phase2 - phase1)


### The functions for computing the coefficients themselves ###

cdef complex_t k_nmnm(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    int n1, int m1,
    int n2, int m2,
) nogil:
    cdef:
        complex_t z1, z2

    if kws_x.aligned and kws_y.aligned:
        if kws_x.matched and kws_y.matched:
            if n1 == n2 and m1 == m2:
                return 1.0 + 0.0j
            else:
                return 0.0 + 0.0j

    z1 = k_mm(kws_x, n1, n2)
    z2 = k_mm(kws_y, m1, m2)

    return z1 * z2

cdef complex_t k_mm(
    const knm_bh_workspace* kws,
    int m1, int m2,
) nogil:
    cdef:
        double out1
        complex_t out2
        complex_t sg, su

        double root_m1f_m2f = sqrt_factorial(m1) * sqrt_factorial(m2)

    # (-1)^{n'}
    out1 = msign(m2)
    # sqrt{n! n'!}
    out2 = root_m1f_m2f
    if not kws.matched:
        out2 *= (
            # (1 + K*)^{-(n + n' + 1) / 2}
            cpow_re(1.0 + kws.K_conj, -0.5 * (m1 + m2 + 1)) *
            # (1 + K0)^{n/2 + 1/4}
            kws.K0_plus1_pow_cache[m1]
        )

    sg = S_g(kws, m1, m2)
    su = S_u(kws, m1, m2)

    if kws.aligned:
        return out1 * out2 * (sg - su)

    # See comments at end of knm_bh_ws_recompute_misalignment for description
    return out1 * kws.ex * crotate2(
        out2 * (sg - su), kws.ex_cos_phase, kws.ex_sin_phase
    )

cdef complex_t S_g(const knm_bh_workspace* kws, int m1, int m2) nogil:
    cdef:
        int s1, s2
        complex_t z, z1, z2
        complex_t nom1, nom2
        double denom1, denom2
        int i, j, k # loop indices

    if m1 == 0 or m1 == 1:
        s1 = 0
    else:
        if m1 % 2: # m1 odd
            s1 = int((m1 - 1) / 2)
        else: # m1 even
            s1 = int(m1 / 2)

    if m2 == 0 or m2 == 1:
        s2 = 0
    else:
        if m2 % 2: # m2 odd
            s2 = int((m2 - 1) / 2)
        else: # m2 even
            s2 = int(m2 / 2)

    z = COMPLEX_0
    for i in range(1 + s1):
        for j in range(1 + s2):
            nom1 = kws.XS_pow_cache[m1 - 2 * i] * kws.X_pow_cache[m2 - 2 * j]
            if nom1 == COMPLEX_0:
                continue

            denom1 = (
                factorial(m1 - 2 * i) *
                factorial(m2 - 2 * j)
            )
            z1 = nom1 * msign(i) / denom1
            z2 = COMPLEX_0
            for k in range(1 + nmin(i , j)):
                nom2 = kws.FS_pow_cache[i - k] * kws.F_pow_cache[j - k]
                if nom2 == COMPLEX_0:
                    continue

                denom2 = (
                    factorial(2 * k) *
                    factorial(i - k) *
                    factorial(j - k)
                )
                z2 += nom2 * msign(k) / denom2

            z += z1 * z2

    return z

cdef complex_t S_u(const knm_bh_workspace* kws, int m1, int m2) nogil:
    cdef:
        int s3, s4
        complex_t z, z1, z2
        complex_t nom1, nom2
        double denom1, denom2
        int i, j, k # loop indices

    if m1 == 0 or m2 == 0:
        return COMPLEX_0

    if m1 == 1 or m1 == 2:
        s3 = 0
    else:
        if m1 % 2: # m1 odd -> (m1 - 1) even
            s3 = int((m1 - 1) / 2)
        else: # m1 even -> (m1 - 1) odd
            s3 = int((m1 - 2) / 2)

    if m2 == 1 or m2 == 2:
        s4 = 0
    else:
        if m2 % 2: # m2 odd -> (m2 -1) even
            s4 = int((m2 - 1) / 2)
        else:
            s4 = int((m2 - 2) / 2)

    z = COMPLEX_0
    for i in range(1 + s3):
        for j in range(1 + s4):
            nom1 = kws.XS_pow_cache[m1 - 2 * i - 1] * kws.X_pow_cache[m2 - 2 * j - 1]
            if nom1 == COMPLEX_0:
                continue

            denom1 = factorial(m1 - 2 * i - 1) * factorial(m2 - 2 * j - 1)
            z1 = nom1 * msign(i) / denom1
            z2 = COMPLEX_0
            for k in range(1 + nmin(i, j)):
                nom2 = kws.FS_pow_cache[i - k] * kws.F_pow_cache[j - k]
                if nom2 == COMPLEX_0:
                    continue

                denom2 = (
                    factorial(2 * k + 1) *
                    factorial(i - k) *
                    factorial(j - k)
                )
                z2 += nom2 * msign(k) / denom2

            z += z1 * z2

    return z



from finesse.cymath cimport complex_t
from finesse.cymath.gaussbeam cimport beam_param
from finesse.cymath.homs cimport unm_workspace, unm_factor_store

from finesse.detectors.workspace cimport (
    DetectorWorkspace,
    MaskedDetectorWorkspace,
    OutputFuncWrapper,
)
from finesse.element cimport BaseCValues
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.frequency cimport frequency_info_t

from finesse.simulations.base cimport NodeBeamParam

cdef enum ScanningAxis:
    XAXIS,
    YAXIS

cdef class CameraWorkspace(MaskedDetectorWorkspace):
    cdef:
        Py_ssize_t node_id

        unm_workspace uws
        unm_factor_store ufs

        double[:, ::1] phase_cache

        const double[::1] x_view
        const double[::1] y_view
        Py_ssize_t xpts
        Py_ssize_t ypts
        ScanningAxis scan_ax

        int nthreads

        const NodeBeamParam* q

    cpdef cache(self, bint initial=?)

### CCD type camera workspaces ###

cdef class CCDWorkspace(CameraWorkspace):
    cdef:
        double[:, ::1] out


cdef class CCDLineWorkspace(CameraWorkspace):
    cdef:
        double[::1] out

# NOTE (sjr) Just here for completion / type checking, doesn't need to store anything
#            extra beyond CameraWorkspace as it just computes a single pixel intensity
cdef class CCDPixelWorkspace(CameraWorkspace):
    pass

### Field / complex camera type workspaces ###

cdef class ComplexCameraValues(BaseCValues):
    cdef public:
        double f

cdef class ComplexCameraWorkspace(CameraWorkspace):
    cdef:
        ComplexCameraValues v

cdef class FieldCameraWorkspace(ComplexCameraWorkspace):
    cdef:
        complex_t[:, ::1] out


cdef class FieldLineWorkspace(ComplexCameraWorkspace):
    cdef:
        complex_t[::1] out

# NOTE (sjr) Again just here for completion / type checking, doesn't need to store anything
#            extra beyond CameraWorkspace as it just computes a single pixel field
cdef class FieldPixelWorkspace(ComplexCameraWorkspace):
    pass



cdef class Frequency:
    cdef:
        str __name
        bint __is_audio, __symbol_changing
        int __order
        int __carrier
        int __index
        object __symbol
        object __lambdified
        double __start_value


"""This struct is used to contain various properties of a frequency bin in
a simulation in a fast C struct. The python side should update this struct
and pass it to C functions, allowing GIL to be released, etc.
"""
cdef struct frequency_info:
    double f # Current frequency bin value [Hz]
    double *f_car # If a sideband, the current carrier frequency bin value [Hz]
    Py_ssize_t index # Index of this frequency bin
    # Optical carrier field members
    Py_ssize_t audio_lower_index # index of lower sideband if signals being simulated
    Py_ssize_t audio_upper_index # index of upper sideband if signals being simulated
    # Optical signal field members
    Py_ssize_t audio_order # audio order, either -1 or +1, 0 means a carrier
    Py_ssize_t audio_carrier_index # index of carrier frequency bin


ctypedef frequency_info frequency_info_t


cdef class FrequencyContainer:
    cdef readonly:
        Py_ssize_t size
        tuple frequencies
    cdef:
        frequency_info_t* frequency_info
        frequency_info_t* carrier_frequency_info

    cdef update_frequency_info(self)
    cdef initialise_frequency_info(self)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport cython

cimport numpy as np
import numpy as np

from finesse.knm.matrix cimport make_unscaled_X_scatter_knm_matrix, make_unscaled_Y_scatter_knm_matrix
from finesse.components.modal.workspace cimport KnmInfo
from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, cexp
from finesse.cymath.gaussbeam cimport bp_beamsize
from finesse.cymath.math cimport exp, sqrt, cos, sin, asin, isfinite
from finesse.cymath.math cimport radians
from finesse.frequency cimport FrequencyContainer
from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray
from finesse.symbols import Symbol
from finesse.simulations.basematrix cimport MatrixSystemSolver, CarrierSignalMatrixSimulation

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.string cimport strcmp, memcpy
from libc.stdlib cimport free, calloc
from finesse.cymath.complex cimport DenseZVector
import logging

ctypedef (double*, double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_9

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD


LOGGER = logging.getLogger(__name__)


cdef class MirrorValues(BaseCValues):
    def __init__(MirrorValues self):
        cdef ptr_tuple_9 ptr = (&self.R, &self.T, &self.L, &self.phi, &self.Rcx, &self.Rcy, &self.xbeta, &self.ybeta, &self.misaligned)
        cdef tuple params = ("R","T","L","phi","Rcx","Rcy","xbeta","ybeta","misaligned")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class MirrorOpticalConnections:
    """Contains C accessible references to submatrices for
    optical connections for this element.
    """
    def __cinit__(self, object mirror, MatrixSystemSolver mtx):
        # Only 1D arrays of submatrices as no frequency coupling happening
        cdef int Nf = mtx.optical_frequencies.size

        self.P1i_P1o = SubCCSView1DArray(Nf)
        self.P2i_P2o = SubCCSView1DArray(Nf)
        self.P1i_P2o = SubCCSView1DArray(Nf)
        self.P2i_P1o = SubCCSView1DArray(Nf)
        self.opt_conn_ptrs.P1i_P1o = <PyObject**>self.P1i_P1o.views
        self.opt_conn_ptrs.P2i_P2o = <PyObject**>self.P2i_P2o.views
        self.opt_conn_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.opt_conn_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class MirrorSignalConnections(MirrorOpticalConnections):
    """Contains C accessible references to submatrices for
    optical, electrical, and mechanical connections for this
    element.
    """
    def __cinit__(self, object mirror, MatrixSystemSolver mtx):
        cdef:
            int Nfo =  mtx.optical_frequencies.size

        Nmz = mirror.mech.z.num_frequencies # num of mechanic frequencies
        self.P1i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P1o_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P2i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P2o_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P1i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P1o_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P2i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P2o_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P1i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P1o_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P2i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P2o_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.Z_P1o = SubCCSView2DArray(Nmz, Nfo)
        self.Z_P2o = SubCCSView2DArray(Nmz, Nfo)
        self.yaw_P1o = SubCCSView2DArray(1, Nfo)
        self.yaw_P2o = SubCCSView2DArray(1, Nfo)
        self.pitch_P1o = SubCCSView2DArray(1, Nfo)
        self.pitch_P2o = SubCCSView2DArray(1, Nfo)

        self.sig_conn_ptrs.P1i_Fz = <PyObject***>self.P1i_Fz.views
        self.sig_conn_ptrs.P2i_Fz = <PyObject***>self.P2i_Fz.views
        self.sig_conn_ptrs.P1o_Fz = <PyObject***>self.P1o_Fz.views
        self.sig_conn_ptrs.P2o_Fz = <PyObject***>self.P2o_Fz.views
        self.sig_conn_ptrs.P1i_Fyaw = <PyObject***>self.P1i_Fyaw.views
        self.sig_conn_ptrs.P2i_Fyaw = <PyObject***>self.P2i_Fyaw.views
        self.sig_conn_ptrs.P1o_Fyaw = <PyObject***>self.P1o_Fyaw.views
        self.sig_conn_ptrs.P2o_Fyaw = <PyObject***>self.P2o_Fyaw.views
        self.sig_conn_ptrs.P1i_Fpitch = <PyObject***>self.P1i_Fpitch.views
        self.sig_conn_ptrs.P2i_Fpitch = <PyObject***>self.P2i_Fpitch.views
        self.sig_conn_ptrs.P1o_Fpitch = <PyObject***>self.P1o_Fpitch.views
        self.sig_conn_ptrs.P2o_Fpitch = <PyObject***>self.P2o_Fpitch.views
        self.sig_conn_ptrs.Z_P1o = <PyObject***>self.Z_P1o.views
        self.sig_conn_ptrs.Z_P2o = <PyObject***>self.Z_P2o.views
        self.sig_conn_ptrs.yaw_P1o = <PyObject***>self.yaw_P1o.views
        self.sig_conn_ptrs.yaw_P2o = <PyObject***>self.yaw_P2o.views
        self.sig_conn_ptrs.pitch_P1o = <PyObject***>self.pitch_P1o.views
        self.sig_conn_ptrs.pitch_P2o = <PyObject***>self.pitch_P2o.views


cdef class MirrorWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, refill):
        cdef FrequencyContainer fcnt
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            MirrorOpticalConnections(owner, sim.carrier),
            MirrorSignalConnections(owner, sim.signal) if sim.signal else None,
            MirrorValues()
        )
        # Casting python objects to known types for faster access
        self.mv = self.values
        self.mcc = self.carrier.connections
        if sim.signal:
            self.mcs = self.signal.connections
        else:
            self.mcs = None

        # These are only use for beam tracing as far as I can tell
        # for the beam tracing, which only has optical nodes.
        self.P1i_id = sim.carrier.node_id(owner.p1.i)
        self.P1o_id = sim.carrier.node_id(owner.p1.o)
        self.P2i_id = sim.carrier.node_id(owner.p2.i)
        self.P2o_id = sim.carrier.node_id(owner.p2.o)

        if sim.signal:
            # If we have a signal simulation then we need to cache some indicies
            # for grabbing data when filling
            self.car_p1o_rhs_idx = sim.carrier.get_node_info(owner.p1.o)['rhs_index']
            self.car_p2o_rhs_idx = sim.carrier.get_node_info(owner.p2.o)['rhs_index']
            self.car_p1i_rhs_idx = sim.carrier.get_node_info(owner.p1.i)['rhs_index']
            self.car_p2i_rhs_idx = sim.carrier.get_node_info(owner.p2.i)['rhs_index']
            self.car_p_num_hom =   sim.carrier.get_node_info(owner.p1.o)['nhoms']

            self.z_signal_enabled = owner.mech.z.full_name in sim.signal.nodes
            if self.z_signal_enabled:
                # Get a reference to the mechanical node frequencies
                fcnt = sim.signal.mechanical_frequencies[owner.mech.z]
                self.z_mech_freqs = fcnt.frequency_info
                self.z_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.z].size

            self.yaw_signal_enabled = owner.mech.yaw.full_name in sim.signal.nodes
            if self.yaw_signal_enabled:
                fcnt = sim.signal.mechanical_frequencies[owner.mech.yaw]
                self.yaw_mech_freqs = fcnt.frequency_info
                self.yaw_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.yaw].size
                self.K_yaw_sig = make_unscaled_X_scatter_knm_matrix(self.sim.model_data.homs_view)

            self.pitch_signal_enabled = owner.mech.pitch.full_name in sim.signal.nodes
            if self.pitch_signal_enabled:
                fcnt = sim.signal.mechanical_frequencies[owner.mech.pitch]
                self.pitch_mech_freqs = fcnt.frequency_info
                self.pitch_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.pitch].size
                self.K_pitch_sig = make_unscaled_Y_scatter_knm_matrix(self.sim.model_data.homs_view)

        self.sym_abcd_Cs[:] = [
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
        ]
        self.abcd_Cs[:] = [
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
        ]

    def __dealloc__(self):
        cdef Py_ssize_t i
        for i in range(8):
            cy_expr_free(self.sym_abcd_Cs[i])

    cpdef compile_abcd_cy_exprs(self):
        cdef:
            object mirror = self.owner
            list abcd_handles = list(mirror._abcd_matrices.values())

        if self.sim.is_modal:
            self.abcd_Cs[:] = [
                <double*>&self.abcd_p1p1_x[1][0], <double*>&self.abcd_p1p1_y[1][0],
                <double*>&self.abcd_p2p2_x[1][0], <double*>&self.abcd_p2p2_y[1][0],
                <double*>&self.abcd_p1p2_x[1][0], <double*>&self.abcd_p1p2_y[1][0],
                <double*>&self.abcd_p2p1_x[1][0], <double*>&self.abcd_p2p1_y[1][0]
            ]

        cdef Py_ssize_t i
        cdef object[:, ::1] M_sym
        for i in range(8):
            M_sym = abcd_handles[i][0]

            # NOTE (sjr) Only element C of a Mirror ABCD matrix can possibly change
            if isinstance(M_sym[1][0], Symbol):
                ch_sym = M_sym[1][0].expand_symbols().eval(keep_changing_symbols=True)
                if isinstance(ch_sym, Symbol):
                    self.sym_abcd_Cs[i] = cy_expr_new()
                    cy_expr_init(self.sym_abcd_Cs[i], ch_sym)

    cpdef update_parameter_values(self):
        ConnectorWorkspace.update_parameter_values(self)

        cdef Py_ssize_t i
        for i in range(8):
            if self.sym_abcd_Cs[i] != NULL and self.abcd_Cs[i] != NULL:
                self.abcd_Cs[i][0] = cy_expr_eval(self.sym_abcd_Cs[i])


cdef inline void mirror_fill_optical_2_optical(
        mirror_optical_connections *conn,
        MirrorWorkspace ws,
        frequency_info_t *freq,
        double r,
        double t,
        double phi_0
    ):
    if ws.mv.misaligned >= 1:
        r = 0

    cdef double phase_shift_scaling = (1 + freq.f / ws.sim.model_data.f0)
    # Phase on reflection is not equal if nr1 != nr2 and AoI != 0
    # so the usual i on transmission phase no longer works.
    cdef double phi_r1, phi_r2, phi_t
    cdef complex_t _r1, _r2, _t0
    if ws.sim.model_data.v2_transmission_phase or ws.nr1 == ws.nr2:
        # old v2 phase on transmission
        # The usual i on transmission and reflections
        # are opposite phase on each side, ignores refractive index
        phi_r1 = 2 * phi_0 * phase_shift_scaling
        _r1 = r * cexp(1j * phi_r1)
        _r2 = conj(_r1)
        _t0 = 1j * t
    else:
        # Uses N=-1, Eq.2.25 in Living Rev Relativ (2016) 19:3 DOI 10.1007/s41114-016-0002-8
        # bs transmission phase depends on the reflectivity, refractive indices,
        # and angle of incidence
        phi_r1 = 2 * phi_0 * ws.nr1 * phase_shift_scaling
        phi_r2 = -2 * phi_0 * ws.nr2 * phase_shift_scaling
        phi_t = PI/2 + 0.5 * (phi_r1 + phi_r2)
        _r1 = r * cexp(1j * phi_r1)
        _r2 = r * cexp(1j * phi_r2)
        _t0 = t * cexp(1j * phi_t)

    # Reflections
    if conn.P1i_P1o[freq.index]:
        (<SubCCSView>conn.P1i_P1o[freq.index]).fill_negative_za_zm_2(
            _r1, ws.K11.mtx.ptr, ws.K11.mtx.stride1, ws.K11.mtx.stride2
        )
    if conn.P2i_P2o[freq.index]:
        (<SubCCSView>conn.P2i_P2o[freq.index]).fill_negative_za_zm_2(
            _r2, ws.K22.mtx.ptr, ws.K22.mtx.stride1, ws.K22.mtx.stride2
        )
    # Transmission
    if conn.P1i_P2o[freq.index]:
        (<SubCCSView>conn.P1i_P2o[freq.index]).fill_negative_za_zm_2(
            _t0, ws.K12.mtx.ptr, ws.K12.mtx.stride1, ws.K12.mtx.stride2
        )
    if conn.P2i_P1o[freq.index]:
        (<SubCCSView>conn.P2i_P1o[freq.index]).fill_negative_za_zm_2(
            _t0, ws.K21.mtx.ptr, ws.K21.mtx.stride1, ws.K21.mtx.stride2
        )


mirror_carrier_fill = FillFuncWrapper.make_from_ptr(c_mirror_carrier_fill)
cdef object c_mirror_carrier_fill(ConnectorWorkspace cws):
    r"""
    Fills the sub-matrix of the interferometer matrix held by `sim`, corresponding
    to the `mirror` component.

    A light field :math:`E_{\mathrm{in}}` reflected by a mirror is in general changed in phase
    and amplitude (in the plane-wave picture):

    .. math::
        E_{\mathrm{refl}} = r\exp{\left(i\varphi\right)} E_{\mathrm{in}},

    where :math:`r` is the amplitude reflectance of the mirror and :math:`\varphi = 2kx` the
    phase shift acquired by the propagation towards and back from the mirror (:math:`x` is the
    displacement from the reference plane - see :numref:`fig_mirror_couplings`).

    .. _fig_mirror_couplings:
    .. figure:: ../images/mirror.*
        :align: center

        Field couplings at mirror with a representation of the reference plane.

    The tuning :math:`\phi` gives the displacement of the mirror expressed in radians, with respect
    to the reference plane. A tuning of :math:`\phi = 2\pi` represents a displacement of the mirror,
    :math:`x_m - x_0` on :numref:`fig_mirror_couplings`, by one carrier wavelength :math:`x = \lambda_0`.

    A certain displacement results in different phase shifts for light fields with different frequencies. The
    phase shift a general field acquires at the reflection on the front surface of the mirror can be written
    as:

    .. math::
        \varphi = 2\phi\frac{\omega}{\omega_0},

    where :math:`omega` is the angular frequency of the reflected light. If a second light beam hits the mirror
    from the other direction, the phase change :math:`\varphi_2` with respect to the same tuning would be:

    .. math::
        \varphi_2 = -\varphi.

    The tuning of a mirror (or beam splitter) does not represent a change in the path length but a change in the
    position of the component. The transmitted light is thus not affected by the tuning of the mirror. Only the
    phase shift of :math:`\pi/2` for every transmission has to be taken into account:

    .. math::
        E_{\mathrm{trans}} = i t E_{\mathrm{in}},

    with :math:`t` as the amplitude transmittance of the mirror.

    Putting all this together, for each frequency light field :math:`f` in the interferometer, the following
    quantities are computed in general (including higher-order spatial modes) for the field couplings,

    .. math::
        \begin{array}{l}
            m_{11} = r K_{11} \exp{\left(i 2\phi \left(1 + \frac{f}{f_0}\right) \right)},\\
            m_{22} = r K_{22} \exp{\left(-i 2\phi \left(1 + \frac{f}{f_0}\right) \right)},\\
            m_{12} = it K_{12},\\
            m_{21} = it K_{21},
        \end{array}

    where :math:`K_{\mathrm{ij}}` are the scattering matrices for each direction (see
    :ref:`scatter_matrices`). Here, each :math:`m_{\mathrm{ij}}` term now represents a
    vector of the couplings of all higher-order spatial mode fields present.
    """
    cdef:
        MirrorWorkspace ws = <MirrorWorkspace>cws
        double t = sqrt(ws.mv.T)
        double r = sqrt(ws.mv.R)
        double phi = radians(ws.mv.phi)
        Py_ssize_t i, size
        mirror_optical_connections *conn = &ws.mcc.opt_conn_ptrs
        frequency_info_t *frequencies

    size = ws.sim.carrier.optical_frequencies.size
    frequencies = ws.sim.carrier.optical_frequencies.frequency_info

    for i in range(size):
        mirror_fill_optical_2_optical(conn, ws, &(frequencies[i]), r, t, phi)


mirror_signal_fill = FillFuncWrapper.make_from_ptr(c_mirror_signal_fill)
cdef object c_mirror_signal_fill(ConnectorWorkspace cws):
    cdef:
        MirrorWorkspace ws = <MirrorWorkspace>cws
        double t = sqrt(ws.mv.T)
        double r = sqrt(ws.mv.R)
        double phi = radians(ws.mv.phi)
        Py_ssize_t i, size
        mirror_optical_connections *car_conn = &ws.mcc.opt_conn_ptrs
        mirror_optical_connections *conn = &ws.mcs.opt_conn_ptrs
        mirror_signal_connections *sconn = &ws.mcs.sig_conn_ptrs
        frequency_info_t *freq
        frequency_info_t *frequencies

    size = ws.sim.signal.optical_frequencies.size
    frequencies = ws.sim.signal.optical_frequencies.frequency_info
    ws.z_to_field = 1j * ws.sim.model_data.k0 * ws.sim.model_data.x_scale
    ws.field_to_F = 1 / (C_LIGHT * ws.sim.model_data.x_scale)

    for i in range(size):
        mirror_fill_optical_2_optical(conn, ws, &(frequencies[i]), r, t, phi)

    if ws.z_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.z_mech_freqs_size == 1:
                single_z_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                multiple_z_mechanical_freq_signal_calc(ws, sconn, car_conn, freq, phi)

    if ws.yaw_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.yaw_mech_freqs_size == 1:
                single_yaw_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                raise NotImplementedError()

    if ws.pitch_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.pitch_mech_freqs_size == 1:
                single_pitch_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                raise NotImplementedError()

cdef void get_carrier_vectors(MirrorWorkspace ws, int carrier_index, DenseZVector *c_p1_i, DenseZVector *c_p2_i, DenseZVector *c_p1_o, DenseZVector *c_p2_o):
    assert(c_p1_i)
    assert(c_p2_i)
    assert(c_p1_o)
    assert(c_p2_o)
    c_p1_i.size = c_p1_o.size = c_p2_i.size = c_p2_o.size = ws.car_p_num_hom
    c_p1_i.stride = c_p1_o.stride = c_p2_i.stride = c_p2_o.stride = 1
    # Get incoming/outgoing carrier field amplitudes
    c_p1_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1i_rhs_idx,
        ws.car_p_num_hom,
        carrier_index,
        0
    )]
    c_p2_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2i_rhs_idx,
        ws.car_p_num_hom,
        carrier_index,
        0
    )]
    c_p1_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1o_rhs_idx,
        ws.car_p_num_hom,
        carrier_index,
        0
    )]
    c_p2_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2o_rhs_idx,
        ws.car_p_num_hom,
        carrier_index,
        0
    )]

cdef void single_yaw_mechanical_frequency_signal_calc (
        MirrorWorkspace ws,
        mirror_signal_connections *conn,
        mirror_optical_connections *car_conn,
        frequency_info_t *freq, # audio sideband
        double phi,
        Py_ssize_t yaw_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        double wx = 0.0
        NodeBeamParam *q_P1o
        NodeBeamParam *q_P2o
        complex_t a_2_o_factor
        complex_t phase_shift = cexp(1j*phi * freq.f / ws.sim.model_data.f0)
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o

    get_carrier_vectors(ws, carrier_index, &c_p1_i, &c_p2_i, &c_p1_o, &c_p2_o)

    # We use an unscaled Knm matrix, so we need to apply the waist size and gouy phase
    # as we always reverse the gouy phase anyway, we just don't bother adding it here
    # k0 scaling with nr is done in code jsut below, along with spot size
    # TODO ddb - these matrix multiplications would be more efficient with a sparse matrix
    # format, CSR maybe? As dense product scales badly with maxtem
    if ws.sim.is_modal: # ignore filling this if doing plane-wave
        a_2_o_factor = 1j * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
        q_P1o = &ws.sim.trace[ws.P1o_id]
        q_P2o = &ws.sim.trace[ws.P2o_id]
        if conn.yaw_P1o[yaw_freq_idx][freq.index]:
            wx = bp_beamsize(&q_P1o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P1o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P1i_P1o[carrier_index]), 0,
                -2/2*ws.nr1 * wx * a_2_o_factor * phase_shift,
                ws.K_yaw_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.yaw_P1o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P2i_P1o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx * a_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        if conn.yaw_P2o[yaw_freq_idx][freq.index]:
            wx = bp_beamsize(&q_P2o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P2o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P2i_P2o[carrier_index]), 0,
                -2/2*ws.nr2 * wx * a_2_o_factor * conj(phase_shift),
                ws.K_yaw_sig.data_view, False
            )
            # Transmission coupling
            (<SubCCSView>conn.yaw_P2o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx * a_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        # -------------------------------------------------
        # Optical to mechanical connections
        # -------------------------------------------------
        # These fill a nHOMx1 matrix to compute RP force
        # There is a minus sign difference between side 1 and 2 here, because
        # of the coordinate system change
        if conn.P1i_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P1i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr1 * wx * ws.field_to_F, &ws.K_yaw_sig.mtx, &c_p1_i
            )
        # differing minus signs here because of the x coordinate flip compared to mechanical node x
        if conn.P1o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P1o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wx * ws.field_to_F, &ws.K_yaw_sig.mtx, &c_p1_o
            )
        if conn.P2i_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P2i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wx * ws.field_to_F, &ws.K_yaw_sig.mtx,  &c_p2_i
            )
        if conn.P2o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P2o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr2 * wx * ws.field_to_F, &ws.K_yaw_sig.mtx, &c_p2_o
            )

cdef void single_pitch_mechanical_frequency_signal_calc (
        MirrorWorkspace ws,
        mirror_signal_connections *conn,
        mirror_optical_connections *car_conn,
        frequency_info_t *freq, # audio sideband
        double phi,
        Py_ssize_t pitch_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        complex_t _tuning  = 0.0
        complex_t _ctuning = 0.0
        double wy          = 0.0
        NodeBeamParam *q_P1o
        NodeBeamParam *q_P2o
        complex_t phase_shift = cexp(1j*phi * freq.f / ws.sim.model_data.f0)
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o

    get_carrier_vectors(ws, carrier_index, &c_p1_i, &c_p2_i, &c_p1_o, &c_p2_o)

    # We use an unscaled Knm matrix, so we need to apply the waist size and gouy phase
    # as we always reverse the gouy phase anyway, we just don't bother adding it here
    # k0 scaling with nr is done in code jsut below, along with spot size
    # TODO ddb - these matrix multiplications would be more efficient with a sparse matrix
    # format, CSR maybe? As dense product scales badly with maxtem
    cdef complex_t a_2_o_factor = 1j * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
    if ws.sim.is_modal:
        q_P1o = &ws.sim.trace[ws.P1o_id]
        q_P2o = &ws.sim.trace[ws.P2o_id]

        if conn.pitch_P1o[pitch_freq_idx][freq.index]:
            wy = bp_beamsize(&q_P1o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.pitch_P1o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P1i_P1o[carrier_index]), 0,
                -2/2 * ws.nr1 * wy * a_2_o_factor * phase_shift,
                ws.K_pitch_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.pitch_P1o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P2i_P1o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wy * a_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )

        if conn.pitch_P2o[pitch_freq_idx][freq.index]:
            wy = bp_beamsize(&q_P2o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            # Extra minus sign here because of coordinate system from back side of of tilted mirror
            (<SubCCSView>conn.pitch_P2o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                # minus because side 2 pitch sends beam upwards
                (<SubCCSView>car_conn.P2i_P2o[carrier_index]), 0, 2/2 * ws.nr2 * wy * a_2_o_factor * conj(phase_shift), ws.K_pitch_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.pitch_P2o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                0.5 * (ws.nr1 - ws.nr2) * wy * a_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )
        # -------------------------------------------------
        # Optical to mechanical connections
        # -------------------------------------------------
        # These fill a nHOMx1 matrix to compute RP force
        # negative sign because a positive displacemtn from side 1 generates a negative pitch torque
        if conn.P1i_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P1i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                - ws.nr1 * wy * ws.field_to_F, &ws.K_pitch_sig.mtx, &c_p1_i
            )
        if conn.P1o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P1o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                - ws.nr1 * wy * ws.field_to_F, &ws.K_pitch_sig.mtx, &c_p1_o
            )
        if conn.P2i_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P2i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy * ws.field_to_F, &ws.K_pitch_sig.mtx, &c_p2_i
            )
        if conn.P2o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P2o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy * ws.field_to_F, &ws.K_pitch_sig.mtx, &c_p2_o
            )


cdef void multiple_z_mechanical_freq_signal_calc (
        MirrorWorkspace ws,
        mirror_signal_connections *conn,
        mirror_optical_connections *car_conn,
        frequency_info_t *freq,
        double phi
    ):
    """Computes the opto-mechanics for a mirror with multiple optical and mechanical frequencies.
    """

    cdef:
        Py_ssize_t i, j
        frequency_info_t *ofrequencies = ws.sim.carrier.optical_frequencies.frequency_info
        Py_ssize_t osize = ws.sim.carrier.optical_frequencies.size
        double fs, fc, fm

    for i in range(osize): # Loop over optical DC
        for j in range(ws.z_mech_freqs_size): # Loop over mechanical frequencies
            fs = freq.f
            fc = ofrequencies[i].f
            fm = ws.z_mech_freqs[j].f

            if (fc-fs == fm) or (fs-fc == fm):
                single_z_mechanical_frequency_signal_calc(
                    ws, conn, car_conn, freq, phi, j, i
                )


cdef void single_z_mechanical_frequency_signal_calc (
        MirrorWorkspace ws,
        mirror_signal_connections *conn,
        mirror_optical_connections *car_conn,
        frequency_info_t *freq,
        double phi,
        Py_ssize_t z_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        complex_t _tuning   = 0.0
        complex_t _ctuning  = 0.0
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o

    get_carrier_vectors(ws, carrier_index, &c_p1_i, &c_p2_i, &c_p1_o, &c_p2_o)

    # -------------------------------------------------
    # Optical to mechanical connections
    # -------------------------------------------------
    # - Longitudinal
    # -------------------------------------------------
    # These fill a nHOMx1 matrix to compute RP force
    if conn.P1i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P1i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field_to_F,
            c_p1_i.ptr, 1, 1
        )
    if conn.P1o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P1o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field_to_F,
            c_p1_o.ptr, 1, 1
        )

    # Minus sign as we force the mirror in the opposite
    # direction from the other side
    if conn.P2i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P2i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field_to_F,
            c_p2_i.ptr, 1, 1
        )
    if conn.P2o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P2o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field_to_F,
            c_p2_o.ptr, 1, 1
        )

    # -----------------------------------------------------------------
    # Mechanical to optical connections
    # -----------------------------------------------------------------
    # - Longitudinal
    # -----------------------------------------------------------------
    # As the output has a mixture of both refl and transmitted we only
    # modulate the incoming and refl'd field so we have to propagate
    # the input

    # As we are using the propagaged carrier, it already has the various phase
    # static phase+amplitude factors, HOM scattering, etc. included, which is
    # useful as that means we don't duplicate the calculations here. However,
    # the phase accumulated is slightly different, as frequency shift happens
    # at the mirror so it picks up a slightly different detuning phase coming
    # back from the mirror, here we correct that
    phase_shift = phi * freq.f / ws.sim.model_data.f0

    # -----------------------------------------------------------------
    # Signal generation z->p1.o
    if conn.Z_P1o[z_freq_idx][freq.index]:
        _tuning = cexp(1.0j * phase_shift)
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P1o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P1i_P1o[carrier_index]), 0, ws.z_to_field * _tuning, False
        )
    # -----------------------------------------------------------------
    # Signal generation z->p2.o
    # extra 180 phase here as we're doing the opposite
    # modulation when looked at from the other side of the mirror
    if conn.Z_P2o[z_freq_idx][freq.index]:
        _ctuning = conj(_tuning)
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P2o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P2i_P2o[carrier_index]), 0, -ws.z_to_field * _ctuning, False
        )


mirror_fill_qnoise = FillFuncWrapper.make_from_ptr(c_mirror_fill_qnoise)
cdef object c_mirror_fill_qnoise(ConnectorWorkspace cws):
    r"""
    Fills the quantum noise input matrix elements corresponding to this `mirror`.
    """
    cdef:
        MirrorWorkspace ws = <MirrorWorkspace> cws
        PyObject ***noises = ws.output_noise.ptrs
        frequency_info_t *freq

        Py_ssize_t i, j

        double qn_internal_loss
        complex_t factor

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        factor = 0.5 * (1 + freq.f_car[0] / ws.sim.model_data.f0)
        qn_internal_loss = ws.mv.L
        if not ws.sim.is_modal:
            (<SubCCSView>noises[0][freq.index]).fill_za(factor * qn_internal_loss)
            (<SubCCSView>noises[1][freq.index]).fill_za(factor * qn_internal_loss)
        else:
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.mv.R * ws.oconn_info[0].loss[j]
                ws.total_losses[j] += ws.mv.T * ws.oconn_info[3].loss[j]

            (<SubCCSView>noises[0][freq.index]).fill_za_dd(factor, ws.total_losses)
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.mv.R * ws.oconn_info[1].loss[j]
                ws.total_losses[j] += ws.mv.T * ws.oconn_info[2].loss[j]

            (<SubCCSView>noises[1][freq.index]).fill_za_dd(factor, ws.total_losses)



ctypedef (double*, ) ptr_tuple_1


cdef class VariableValues(BaseCValues):
    def __init__(VariableValues self):
        cdef ptr_tuple_1 ptr = (&self.value, )
        cdef tuple params = ("value", )
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class VariableWorkspace(ElementWorkspace):
    def __init__(self, owner):
        super().__init__(owner, VariableValues())
        self.vv = self.values




cdef class TraceForest:
    cdef public:
        bint symmetric # Should beam traces on this forest set opposite node q to -q*?

    cdef readonly:
        list forest
        Py_ssize_t N_trees

        list dependencies # the Cavities and Gausses in order of tracing priority

        object model

    cdef:
        list node_coverage
        list internal_cavity_nodes

    cpdef void clear(self)

    cdef int _add_internal_cavity_trees(self, list cavities) except -1
    cdef int _handle_overlapping_cavities(self) except -1
    cdef int _add_external_cavity_tree(self, object cav) except -1
    cdef int _add_gauss_tree(self, object gauss) except -1

    cdef int _add_beamsplitter_branch_trees(self) except -1
    cdef int _add_backwards_nonsymm_trees(self) except -1

    cpdef set find_untraversed_nodes(self)

    cpdef TraceTree find_tree_from_node(self, object node)
    cpdef object find_dependency_from_node(self, object node, bint raise_not_found=?)
    cpdef set gather_nodes(self)

    cdef object node_from_name(self, name)

    cdef TraceForest make_changing_forest(self)

    cpdef bint contains_space(self, object space)
    cpdef bint contains_comp(self, object comp)

    cpdef tuple find_potential_mismatch_couplings(self, TraceForest other=?)
    cpdef tuple find_intersection_couplings(self, TraceForest other=?)
    cpdef tuple get_mirror_reflection_couplings(
        self,
        bint ignore_internal_cavities=?,
        list skip_dependencies=?,
    )

    cdef void trim(self)

    cpdef draw_by_dependency(self)
    cpdef draw(self)


cdef class TraceTree:
    cdef readonly:
        TraceTree parent
        TraceTree left
        TraceTree right

        # Either a Cavity or a Gauss
        object dependency
        object node

        bint is_x_changing
        bint is_y_changing

        # Is this tree an internal cavity representation or directly
        # froma Gauss command node?
        bint is_source

        # Numeric views on the component ABCD matrices - the initial
        # numeric ABCD memory for each component is only ever modified,
        # never re-allocated, so these views will always be valid
        double[:, ::1] left_abcd_x
        double[:, ::1] left_abcd_y
        double[:, ::1] right_abcd_x
        double[:, ::1] right_abcd_y
        # Is the left tree a reflection from a surface? Important for
        # computing composite / round-trip ABCDs so co-ordinate transformation
        # in x plane can be taken into account
        bint is_left_surf_refl

        # Symbolic ABCD views - only to be used for utility beam tracing code
        # such as computing a composite ABCD matrix over a path
        object[:, ::1] sym_left_abcd_x
        object[:, ::1] sym_left_abcd_y
        object[:, ::1] sym_right_abcd_x
        object[:, ::1] sym_right_abcd_y

        double nr

    cdef:
        # Indices of the node (and opposite for symmetric tracing) to be
        # set and used by a BaseSimulation
        Py_ssize_t node_id
        Py_ssize_t opp_node_id

        # Only used for non-symmetric traces on trees which couldn't be
        # reached from any other dependency branch. This flag indicates
        # the next (left) tree node is a reverse coupling so need to
        # apply the inverse abcd law transform when tracing.
        bint do_inv_transform

        # Again, only used for non-symmetric traces on trees which couldn't
        # be reached from any other dependency branch AND for which the node
        # doesn't have a reverse coupling; e.g. reverse node at a Laser. This
        # flag indicates that the next tree node needs -q* applied during tracing
        # as there is really no other option for setting it at this point.
        bint do_nonsymm_reverse

    @staticmethod
    cdef TraceTree initialise(
        object node, object dependency, bint* is_dependency_changing=?
    )

    cpdef TraceTree add_left(self, TraceTree sub_tree)
    cpdef TraceTree add_right(self, TraceTree sub_tree)

    cpdef bint remove_left(self)
    cpdef bint remove_right(self)

    cdef void trim_at_nodes(
        self, list nodes, bint include_opposite=?
    )

    cdef TraceTree _propagate_find_tree_at_node(
        self,
        TraceTree tree,
        object node,
        bint include_opposite,
    )
    cpdef TraceTree find_tree_at_node(self, object node, bint include_opposite=?)

    cdef _get_last_input_nodes(self, list last_nodes)
    cpdef list get_last_input_nodes(self)

    cpdef bint contains_node(self, object node)
    cpdef bint contains_space(self, object space)
    cpdef bint contains_comp(self, object comp)

    cdef __append_mirror_refl_coupling(self, list couplings)
    cdef _get_mirror_refl_couplings(self, list couplings)
    cpdef list get_mirror_reflection_couplings(self)

    @staticmethod
    cdef bint subtrees_changing(TraceTree sub_tree)
    cpdef bint is_changing(self, bint recursive=?)

    cdef list get_broadest_changing_subtrees(self)
    cpdef draw(self, unicode left_pad=?)

    cpdef trace_beam(self, dict trace, double lambda0, bint symmetric)


cpdef TraceTree get_last_left_branch(TraceTree tree)

cpdef bint is_abcd_changing(object[:, ::1] M)

cpdef void update_rt_abcd(
    TraceTree tree,
    double[:, ::1] ABCDx,
    double[:, ::1] ABCDy
)



# cython profile=True

import logging
import numpy as np
import gc

from finesse.config import config_instance
from finesse.cmatrix import KLUMatrix
from finesse.cmatrix cimport KLUMatrix
from finesse.cymath cimport complex_t
from finesse.simulations.basematrix cimport MatrixSystemSolver

LOGGER = logging.getLogger(__name__)

from finesse.components.modulator import Modulator
from finesse.components.modal.modulator cimport ModulatorWorkspace
from finesse.components.general import NoiseType
from finesse.detectors.general import Detector, NoiseDetector
from finesse.detectors.workspace import DetectorWorkspace
from finesse.components.workspace cimport ConnectorWorkspace, fill_list_t

class KLUStatWorkspace(DetectorWorkspace):
    pass

class KLUStatsDetector(Detector):
    """
    A detector that can be used to output some of the KLU solver statistics.
    """

    def __init__(self, name, matrix, stat):
        Detector.__init__(self, name, None, dtype=np.float64, label="a.u")
        self.stat = stat.lower()
        self.matrix = matrix.lower()
        if matrix.lower() not in ('carrier', 'signal'):
            raise Exception("Matrix setting must be either `carrier` or `signal`")
        if stat.lower() not in ('rcond', 'growth', 'condest'):
            raise Exception("Matrix setting must be either 'rcond', 'growth', or 'condest'")

    def _get_workspace(self, sim):
        ws = KLUStatWorkspace(self, sim)
        ws.fn = getattr(sim, self.stat)
        ws.set_output_fn(self.fill)
        return ws

    def fill(self, ws):
        return ws.fn()


cdef class KLUMatrixSystemSolver(MatrixSystemSolver):
    cdef double prev_rcond
    cdef double rcond_diff_lim

    def __init__(self, *args, **kwargs):
        super().__init__(KLUMatrix, *args, **kwargs)
        self.rcond_diff_lim = float(config_instance()['klu']['rcond_diff_lim'])

    cdef refactor(self):
        #cdef double rcond = (<KLUMatrix>self._M).rcond()
        #cdef double diff = abs((self.prev_rcond - rcond)/self.prev_rcond)

        #if diff > self.rcond_diff_lim:
        #    self._M.factor()
        #    self.prev_rcond = rcond
        #else:
        #    self._M.refactor()
        self._M.refactor()

    cdef solve(self):
        self._M.solve()
        #self._M.print_rhs()
        #self._M.print_matrix()

    cdef solve_noises(self):
        cdef:
            Py_ssize_t idx, i
            bint factor = False
            KLUMatrix noise_M
            complex_t[::1] rhs
            fill_list_t *fill_list
            ConnectorWorkspace cws

        # Modulators couple things differently for the purposes of quantum noise, so we need to
        # refill them here
        # TODO: Really this shouldn't be needed, and the modulator should just behave as normal.
        # However, we don't currently model the extra sidebands required, and so would lose some
        # quantum noise if we were to do so. Instead, we just treat it as if the modulator were not
        # there - this is correct as long as squeezing is not involved.
        if NoiseType.QUANTUM in self._noise_matrices:
            for ws in self.workspaces.to_initial_fill:
                if isinstance(ws.owner, Modulator):
                    ws.fill_quantum_matrix()
                    factor = True
            if factor:
                self.refactor()

        # First ask all of the noise detectors to fill their selection vectors
        for ws in self.workspaces.noise_detectors:
            ws.fill_selection_vector()

        # Then solve all of these at once
        self._M.solve_extra_rhs(transpose=True, conjugate=True)

        # Next, perform the necessary multiplication for each solved rhs
        # TODO: This should be parallelizable per-noise-source, if we can guarantee that the rhs
        # vectors of the same noise type are contiguous
        for ws in self.workspaces.noise_detectors:
            noise_M = self._noise_matrices[ws.owner.noise_type]
            rhs_view = noise_M.get_rhs_view(0)
            rhs_array = np.asarray(rhs_view)
            for idx in ws.owner._requested_selection_vectors.values():
                rhs = self._M.get_rhs_view(idx)
                if ws.owner._has_sources():
                    source_s = ws.get_source_selection_vector()
                    np.multiply(rhs, source_s, out=rhs_array)
                else:
                    rhs_view[:] = rhs[:]
                noise_M.zgemv(rhs)

        # Solve all of the rhs vectors at once again
        self._M.solve_extra_rhs(transpose=False, conjugate=False)

        # Finally give the results back to the noise detectors
        for ws in self.workspaces.noise_detectors:
            for name, idx in ws.owner._requested_selection_vectors.items():
                rhs = self._M.get_rhs_view(idx)
                ws.set_covariance_matrix(rhs, name)

        # Restore the original signal matrix if it's not going to be refilled on the next step
        # TODO: is this needed? Probably
        if NoiseType.QUANTUM in self._noise_matrices:
            for cws in self.workspaces.to_initial_fill:
                if isinstance(cws.owner, Modulator) and cws not in self.workspaces.to_refill:
                    fill_list  = &cws.signal.matrix_fills
                    for i in range(fill_list.size):
                        if fill_list.infos[i].fn_c:
                            fill_list.infos[i].fn_c(cws)
                        elif fill_list.infos[i].fn_py:
                            (<object>fill_list.infos[i].fn_py).__call__(cws)
            if factor:
                self.refactor()

    cpdef double rcond(self):
        return (<KLUMatrix>self._M).rcond()

    cpdef double condest(self):
        return (<KLUMatrix>self._M).condest()

    cpdef double rgrowth(self):
        return (<KLUMatrix>self._M).rgrowth()

    cdef construct(self):
        # Initialising the simulation expects there to be a self._M class that handles the
        # matrix build/memory/etc. This must be set before initialising.
        self._M.construct()
        if self.is_signal_matrix:
            for M in self._noise_matrices.values():
                M.construct(diagonal_fill=0)
        # Point output of matrix to RHS view which klu replaces with solution after solving
        self.out_view = self._M.rhs_view[0]

    cdef initial_run(self):
        self.initial_fill()
        self._M.factor()
        self.prev_rcond = self._M.rcond()
        self.fill_rhs()
        self.solve()

    cdef add_noise_matrix(self, object key):
        if key in self._noise_matrices:
            raise ValueError(f"Noise matrix '{key}' already added to system.")
        self._noise_matrices[key] = KLUMatrix(str(key))

    cdef add_rhs(self, unicode key):
        self._M.get_rhs_index(key)

    # def __enter__(self):
    #     """
    #     When entered the Simulation object will create the matrix to be used in
    #     the simulation.

    #     This is where the matrix gets allocated and constructed. It will expect
    #     that the model structure does not change after this until it has been
    #     exited.
    #     """
    #     # Set the matrix method we'll be using to run this simulation
    #     self._M = KLUMatrix(self.name)
    #     # Initialising the simulation expects there to be a self._M class that handles the
    #     # matrix build/memory/etc. This must be set before initialising.

    #     self._initialise()

    #     self._M.construct()
    #     # Point output to RHS view which klu replaces with solution after solving
    #     self.out_view = self._M.rhs_view

    #     self.initial_fill()
    #     self._M.factor()
    #     self.prev_rcond = self._M.rcond()
    #     self.rcond_diff_lim = float(config_instance()['klu']['rcond_diff_lim'])
    #     self.fill_rhs()
    #     self.solve()

    #     return self



from libc.stdlib cimport free, calloc

from finesse.cmatrix cimport SubCCSView1DArray
from finesse.knm cimport KnmMatrix
from finesse.components.workspace cimport ConnectorWorkspace
from finesse.element cimport BaseCValues
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.components.modal.workspace cimport KnmConnectorWorkspace

from finesse.cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)

from cpython.ref cimport PyObject


cdef class LensValues(BaseCValues):
    cdef public:
        double f


cdef struct lens_connections:
    PyObject** P1i_P2o
    PyObject** P2i_P1o


cdef class LensConnections:
    cdef public:
        int P1i_P2o_idx
        int P2i_P1o_idx
    cdef readonly:
        SubCCSView1DArray P1i_P2o
        SubCCSView1DArray P2i_P1o
    cdef:
        lens_connections conn_ptrs


cdef class LensWorkspace(KnmConnectorWorkspace):
    cdef public:
        # Complete scattering matrices for each propagation direction
        KnmMatrix K12
        KnmMatrix K21

        # Arrays of scattering losses for each mode coupling
        double[::1] K12_loss
        double[::1] K21_loss

        # Refractive indices of adjacent spaces
        double nr1
        double nr2

        # ABCD matrix views (note same for forward & backward propagations)
        double[:, ::1] abcd_x
        double[:, ::1] abcd_y
    cdef:
        LensValues lv
        LensConnections lc

        # Indices (in sim._c_node_info and sim.trace) of nodes
        Py_ssize_t P1i_id
        Py_ssize_t P1o_id
        Py_ssize_t P2i_id
        Py_ssize_t P2o_id

        # Only elements C of ABCD matrices can change, store pointers
        # to compiled changing expression for these elements here
        cy_expr* sym_abcd_Cx
        cy_expr* sym_abcd_Cy



from finesse.frequency cimport frequency_info_t
from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.cymath cimport complex_t
from finesse.simulations.base cimport ModelData, NodeBeamParam
from finesse.element cimport BaseCValues
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.components.modal.workspace cimport KnmConnectorWorkspace
from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np


cdef class OpticalBandpassValues(BaseCValues):
    cdef public:
        double fc
        double bandwidth


cdef struct optical_bandpass_connections:
    PyObject** P1i_P2o
    PyObject** P2i_P1o


cdef class OpticalBandpassConnections:
    cdef public:
        int P1i_P2o_idx
        int P2i_P1o_idx

    cdef readonly:
        SubCCSView1DArray P1i_P2o
        SubCCSView1DArray P2i_P1o
    cdef:
        optical_bandpass_connections ptrs


cdef class OpticalBandpassWorkspace(KnmConnectorWorkspace):
    cdef public:
        KnmMatrix K12
        KnmMatrix K21
        double[::1] K12_loss
        double[::1] K21_loss
        double nr1
        double nr2
    cdef:
        OpticalBandpassValues cvalues
        OpticalBandpassConnections signal_opt_conns
        OpticalBandpassConnections carrier_opt_conns


from enum import Flag
from itertools import combinations
import warnings
import abc

import numpy as np

from finesse import constants
from finesse.utilities import is_iterable

import operator
import logging


LOGGER = logging.getLogger(__name__)

MAKE_LOP = lambda name, opfn: lambda self, other: Operation(name, opfn, self, other)
MAKE_ROP = lambda name, opfn: lambda self, other: Operation(name, opfn, other, self)


# Supported operators.
OPERATORS = {
    "__add__": MAKE_LOP("+", operator.add),
    "__sub__": MAKE_LOP("-", operator.sub),
    "__mul__": MAKE_LOP("*", operator.mul),
    "__radd__": MAKE_ROP("+", operator.add),
    "__rsub__": MAKE_ROP("-", operator.sub),
    "__rmul__": MAKE_ROP("*", operator.mul),
    "__pow__": MAKE_LOP("**", operator.pow),
    "__truediv__": MAKE_LOP("/", operator.truediv),
    "__rtruediv__": MAKE_ROP("/", operator.truediv),
    "__floordiv__": MAKE_LOP("//", operator.floordiv),
    "__rfloordiv__": MAKE_ROP("//", operator.floordiv),
    "__matmul__": MAKE_LOP("@", operator.matmul),
}

# Built-in functions.
FUNCTIONS = {
    "abs": lambda x: Operation("abs", operator.abs, x),
    "neg": lambda x: Operation("neg", operator.neg, x),
    "pos": lambda x: Operation("pos", operator.pos, x),
    "conj": lambda x: Operation("conj", np.conj, x),
    "real": lambda x: Operation("real", np.real, x),
    "imag": lambda x: Operation("imag", np.imag, x),
    "exp": lambda x: Operation("exp", np.exp, x),
    "sin": lambda x: Operation("sin", np.sin, x),
    "arcsin": lambda x: Operation("arcsin", np.arcsin, x),
    "cos": lambda x: Operation("cos", np.cos, x),
    "arccos": lambda x: Operation("arccos", np.arccos, x),
    "tan": lambda x: Operation("tan", np.tan, x),
    "arctan2": lambda y, x: Operation("arctan2", np.arctan2, y, x),
    "sqrt": lambda x: Operation("sqrt", np.sqrt, x),
    "radians": lambda x: Operation("radians", np.radians, x),
    "degrees": lambda x: Operation("degrees", np.degrees, x),
    "deg2rad": lambda x: Operation("deg2rad", np.deg2rad, x),
    "rad2deg": lambda x: Operation("rad2deg", np.rad2deg, x),
    "linspace": lambda a, b, c: Operation("linspace", np.linspace, a, b, c),
    "logspace": lambda a, b, c: Operation("logspace", np.logspace, a, b, c),
    "geomspace": lambda a, b, c: Operation("geomspace", np.geomspace, a, b, c),
}

op_repr = {
    operator.add: "({}+{})",
    operator.sub: "({}-{})",
    operator.mul: "{}*{}",
    operator.pow: "{}**{}",
    operator.truediv: "{}/{}",
    operator.floordiv: "{}//{}",
    operator.mod: "({}%{})",
    operator.matmul: "({}@{})",
    operator.neg: "-{}",
    operator.pos: "+{}",
    operator.abs: "|{}|",
    np.conj: "conj({})",
    np.sqrt: "sqrt({})",
}


def finesse2sympy(expr, iter_num=0):
    """"""
    import sympy
    from finesse.parameter import ParameterRef
    iter_num += 1
    if isinstance(expr, Constant):
        return expr.value
    elif isinstance(expr, ParameterRef):
        return sympy.Symbol(expr.name)
    elif isinstance(expr, Operation):
        sympy_args = [finesse2sympy(arg,iter_num) for arg in expr.args]
        if expr.op == operator.mul:
            op = sympy.Mul
        elif expr.op == operator.add:
            op = sympy.Add
        elif expr.op == operator.sub:
            op = lambda x,y: sympy.Add(x,-y)
        elif expr.op == np.conj:
            op = sympy.conjugate
        elif expr.op == np.radians:
            op = sympy.rad
        elif expr.op == np.exp:
            op = sympy.exp
        elif expr.op == np.sqrt:
            op = sympy.sqrt
        elif expr.op == operator.neg:
            op = lambda x: sympy.Mul(-1, x)
        else:
            raise Exception(f"undefined Operation {expr.op} in {expr}")
        return op(*sympy_args)
    else:
        raise Exception(f'{expr} undefined')

def sympy2finesse(expr, symbol_dict={}, iter_num=0):
    import sympy
    iter_num += 1
    if isinstance(expr,sympy.Mul):
        return np.product([sympy2finesse(arg, symbol_dict, iter_num=iter_num) for arg in expr.args])
    elif isinstance(expr,sympy.Add):
        return np.sum([sympy2finesse(arg, symbol_dict, iter_num=iter_num) for arg in expr.args])
    elif isinstance(expr,sympy.conjugate):
        return np.conj(sympy2finesse(*expr.args, symbol_dict))
    elif isinstance(expr,sympy.exp):
        return np.exp(sympy2finesse(*expr.args, symbol_dict))
    elif isinstance(expr,sympy.Pow):
        return np.power(sympy2finesse(expr.args[0], symbol_dict), sympy2finesse(expr.args[1], symbol_dict))
    elif expr.is_NumberSymbol: # sympy class for named symbols (eg Pi, golden ratio, ...)
        if str(expr) == 'pi':
            return CONSTANTS['pi']
        else:
            return complex(expr)
    elif expr.is_number:
        if expr.is_integer:
            return int(expr)
        elif expr.is_real:
            return float(expr)
        else:
            return complex(expr)
    elif expr.is_symbol:
        return symbol_dict[str(expr)]
    else:
        raise Exception(f'{expr} undefined')


def as_symbol(x):
    return x if isinstance(x, Symbol) else Constant(x)


def display(a):
    """
    For a given Symbol this method will return a human readable string
    representing the various operations it contains.

    Parameters
    ----------
    a : :class:`.Symbol`
        Symbol to print

    Returns
    -------
    String form of Symbol
    """
    if hasattr(a, "op"):
        # Check if operation has a predefined string format
        if a.op in op_repr:
            sop = op_repr[a.op]
        else:  # if not just treat it as a function
            sop = a.op.__name__ + "(" + ("{}," * len(a.args)).rstrip(",") + ")"

        sargs = (display(_) for _ in a.args)

        return sop.format(*sargs)
    elif hasattr(a, "name"):  # Anything with a name attribute just display that
        return a.name
    elif type(a) is Symbol:
        return f"<Symbol @ {hex(id(a))}>"
    else:
        return str(a)


def evaluate(x):
    """Evaluates a symbol or N-dimensional array of symbols.

    Parameters
    ----------
    x : :class:`.Symbol` or array-like
        A symbolic expression or an array of symbolic expressions.

    Returns
    -------
    out : float, complex, :class:`numpy.ndarray`
        A single value for the evaluated expression if `x` is not
        array-like, otherwise an array of the evaluated expressions.
    """
    if is_iterable(x):
        y = np.array(x, dtype=np.complex128)
        if not np.any(y.imag): # purely real symbols in array
            with warnings.catch_warnings():
                # suppress 'casting to float discards imag part' warning
                # as we know that all imag parts are zero here anyway
                warnings.simplefilter("ignore", category=np.ComplexWarning)
                y = np.array(y, dtype=np.float64)

        return y

    if isinstance(x, Symbol):
        return x.eval()

    # If not a symbol then just return x directly
    return x


class Symbol(abc.ABC):
    __add__ = OPERATORS["__add__"]
    __sub__ = OPERATORS["__sub__"]
    __mul__ = OPERATORS["__mul__"]
    __radd__ = OPERATORS["__radd__"]
    __rsub__ = OPERATORS["__rsub__"]
    __rmul__ = OPERATORS["__rmul__"]
    __pow__ = OPERATORS["__pow__"]
    __truediv__ = OPERATORS["__truediv__"]
    __rtruediv__ = OPERATORS["__rtruediv__"]
    __floordiv__ = OPERATORS["__floordiv__"]
    __rfloordiv__ = OPERATORS["__rfloordiv__"]
    __matmul__ = OPERATORS["__matmul__"]

    __abs__ = FUNCTIONS["abs"]
    __neg__ = FUNCTIONS["neg"]
    __pos__ = FUNCTIONS["pos"]
    conjugate = FUNCTIONS["conj"]
    conj = FUNCTIONS["conj"]
    real = FUNCTIONS["real"]
    imag = FUNCTIONS["imag"]
    exp = FUNCTIONS["exp"]
    sin = FUNCTIONS["sin"]
    arcsin = FUNCTIONS["arcsin"]
    cos = FUNCTIONS["cos"]
    arccos = FUNCTIONS["arccos"]
    tan = FUNCTIONS["tan"]
    sqrt = FUNCTIONS["sqrt"]
    radians = FUNCTIONS["radians"]
    degrees = FUNCTIONS["degrees"]
    deg2rad = FUNCTIONS["deg2rad"]
    rad2deg = FUNCTIONS["rad2deg"]

    def __float__(self):
        v = self.eval()
        if np.isscalar(v):
            return float(v)
        else:
            raise TypeError(f"Can't cast {type(v)} into a single float value")

    def __complex__(self):
        v = self.eval()
        if np.isscalar(v):
            return complex(v)
        else:
            raise TypeError(f"Can't cast {type(v)} into a single complex value")

    def __int__(self):
        v = self.eval()
        if np.isscalar(v):
            return int(v)
        else:
            raise TypeError(f"Can't cast {type(v)} into a single int value")

    @property
    def value(self):
        return self.eval()

    def __str__(self):
        return display(self)

    def __repr__(self):
        return f"<Symbolic='{display(self)}' @ {hex(id(self))}>"

    @property
    def is_changing(self):
        """
        Returns True if one of the arguements of this symbolic object
        is varying whilst a :class:`` is running.
        """
        res = False

        if hasattr(self, "op"):
            res = any([_.is_changing for _ in self.args])
        elif hasattr(self, "parameter"):
            res = self.parameter.is_tunable or self.parameter.is_changing

        return res

    def parameters(self, memo=None):
        """
        Returns all the parameters that are present in this symbolic statement
        """
        if memo is None:
            memo = set()

        if hasattr(self, "op"):
            for _ in self.args:
                _.parameters(memo)
        elif hasattr(self, "parameter"):
            memo.add(self)

        return list(memo)

    def changing_parameters(self):
        p = np.array(self.parameters())
        return list(p[list(map(lambda x: x.is_changing, p))])

    def to_sympy(self):
        return finesse2sympy(self)

    def sympy_simplify(self):
        """Converts this expression into a Sympy symbol
        """
        refs = {_.name: _ for _ in self.parameters()} # get a list of symbols we're using
        sympy = finesse2sympy(self)
        return sympy2finesse(sympy.simplify(), refs)

    def expand_symbols(self):
        """A method that expands any symbolic parameter references that are themselves
        symbolic. This can be used to get an expression that only depends on references
        that are numeric.

        Examples
        --------
        >>> import finesse
        >>> model = finesse.Model()
        >>> model.parse(
        ...     '''
        ...     var d 300
        ...     var c 6000
        ...     var b &c+&d
        ...     var a &b+1
        ...     '''
        ... )
        >>> model.a.value.value.expand_symbols()
        <Symbolic='((c.value+d.value)+1)' @ 0x7faa4d351c10>

        Parameters
        ----------
        sym : Symbolic
            Symbolic equation to expand
        """
        def process(p):
            if p.parameter.is_symbolic:
                return p.parameter.value
            else:
                return p

        def _expand(sym):
            if not all(p.parameter.is_symbolic for p in  sym.parameters()):
                return None
            else:
                return sym.eval(subs={p: process(p) for p in sym.parameters()})

        sym = self
        while True:
            res = _expand(sym)
            if res is None:
                return sym
            else:
                sym = res

class Constant(Symbol):
    """Defines a constant symbol that can be used in symbolic math.

    Parameters
    ----------
    value : float, int
        Value of constant

    name : str, optional
        Name of the constant to use when printing
    """

    def __init__(self, value, name=None):
        self.__value = value
        self.__name = name

    def __str__(self):
        return self.__name or str(self.__value)

    def __repr__(self):
        return str(self.__name or self.__value)

    def __eq__(self, obj):
        if isinstance(obj, Constant):
            return obj.value == self.value
        else:
            return False

    def __hash__(self):
        """Constant hash.

        This is used by the tokenizer. Constants are by definition immutable.
        """
        # Add the class to reduce chance of hash collisions.
        return hash((type(self), self.value))

    def eval(self, **kwargs):
        return self.__value

# Constants.
# NOTE: The keys here are used by the parser to recognise constants in kat script.
CONSTANTS = {
    "pi": Constant(constants.PI, name="π"),
    "c0": Constant(constants.C_LIGHT, name="c"),
}


class Resolving(Symbol):
    """A special symbol that represents a symbol that is not yet resolved.

    This is used in the parser to support self-referencing parameters.

    An error is thrown if the value is attempted to be read.
    """
    def eval(self, **kwargs):
        raise RuntimeError(
            "an attempt has been made to read the value of a resolving symbol (hint: symbols "
            "should not evaluated until parsing has fully finished)"
        )

    @property
    def name(self):
        return "RESOLVING"


class Variable(Symbol):
    """
    Makes a variable symbol that can be used in symbolic math. Values must be substituted
    in when evaluating an expression.

    Examples
    --------
    Using some variables to make an expression and evaluating it:

    >>> import numpy as np
    >>> x = Variable('x')
    >>> y = Variable('y')
    >>> z = 4*x**2 - np.cos(y)
    >>> print(f"{z} = {z.eval(subs={x:2, y:3})} : x={2}, y={3}")
    (4*x**2-y) = 13 : x=2, y=3

    Parameters
    ----------
    value : float, int
        Value of constant

    name : str, optional
        Name of the constant to use when printing
    """

    def __init__(self, name):
        if name is None:
            raise ValueError("Name must be provided")
        self.__name = str(name)

    @property
    def name(self):
        return self.__name

    def eval(self, subs=None, **kwargs):
        if subs:
            if self in subs:
                return subs[self]

        return self


class Operation(Symbol):
    """
    This is a symbol to represent a mathematical operation. This could be
    a simple addition, or a more complicated multi-argument function.

    It supports creating new mathematical operations::

        import math
        import cmath

        cos   = lambda x: finesse.symbols.Operation("cos", math.cos, x)
        sin   = lambda x: finesse.symbols.Operation("sin", math.sin, x)
        atan2 = lambda y, x: finesse.symbols.Operation("atan2", math.atan2, y, x)

    Complex math can also be used::

        import numpy as np
        angle = lambda x: finesse.symbols.Operation("angle", np.angle, x)
        print(f"{angle(1+1j)} = {angle(1+1j).eval()}")

    The equality operator is overridden to provide a very basic
    symbolic equality test. It only tests whether two symbolic
    statements are *exactly* the same, e.g.:

        >>> y+x == y+x # True
        >>> y+x == x+y # False

    This could be fixed by making operators more

    Parameters
    ----------
    name : str
        The operation name. This is used for dumping operations to kat script.

    operation : callable
        The function to pass the arguments of this operation to.

    Other Parameters
    ----------------
    *args
        The arguments to pass to `operation` during a call.
    """

    def __init__(self, name, operation, *args):
        self.name = str(name)
        self.args = tuple(as_symbol(_) for _ in args)
        self.op = operation

    def eval(self, **kwargs):
        """Evaluates the operation.

        Parameter substitutions can be given via an optional
        ``subs`` dict (mapping parameters to substituted values).

        Returns
        -------
        result : number or array-like
            The single-valued result of evaluation of the operation (if no
            substitutions given, or all substitutions are scalar-valued). Otherwise,
            if any parameter substitution was a :class:`numpy.ndarray`, then a corresponding array
            of results.
        """
        return self.op(*(_.eval(**kwargs) for _ in self.args))

    def __eq__(self, obj):
        if isinstance(obj, Operation):
            if self.op == obj.op:
                return all([a == b for a, b in zip(self.args, obj.args)])
            else:
                return False
        else:
            return False



cimport numpy as np

cdef extern from "complex.h" nogil:
    double cabs(double complex z)
    double carg(double complex z)
    double complex cexp(double complex z)
    double cimag(double complex z)
    double creal(double complex z)
    double complex csqrt(double complex z)
    double complex conj(double complex z)

cdef extern from "math.h" nogil:
    double cos(double arg)
    double sin(double arg)

ctypedef np.complex128_t complex_t

"""Dense complex matrix, contiguous in memory"""
cdef struct DenseZMatrix:
    complex_t *ptr
    Py_ssize_t stride1 # in units of 16 bytes
    Py_ssize_t stride2 # in units of 16 bytes
    Py_ssize_t size1
    Py_ssize_t size2

"""Dense complex vector, contiguous in memory"""
cdef struct DenseZVector:
    complex_t *ptr
    Py_ssize_t stride # in units of 16 bytes
    Py_ssize_t size

cpdef inline double cnorm(complex_t z) nogil:
    return creal(z) * creal(z) + cimag(z) * cimag(z)


cpdef bint ceq(complex_t z1, complex_t z2) nogil


cpdef inline complex_t inverse_unsafe(complex_t z) nogil:
    cdef double inv_abs_sqd_z = 1.0 / cnorm(z)
    return creal(z) * inv_abs_sqd_z - 1.0j * cimag(z) * inv_abs_sqd_z


cpdef complex_t cpow_re(complex_t z, double n) nogil


cpdef inline complex_t crotate(complex_t z, double ph) nogil:
    cdef:
        double cph = cos(ph)
        double sph = sin(ph)

    return crotate2(z, cph, sph)

cpdef inline complex_t crotate2(complex_t z, double cph, double sph) nogil:
    cdef:
        double zre = creal(z)
        double zim = cimag(z)

    return zre * cph - zim * sph + 1j * (zre * sph + zim * cph)


# TODO (sjr) Move to new finesse.cymath.linalg file and use
#            fused types to make it like a template function
cpdef void couter(complex_t[::1] x, complex_t[::1] y, complex_t[:, ::1] out) nogil




from finesse.cymath.complex cimport complex_t

cdef extern from "constants.h":
    double complex COMPLEX_0 # 0 + 0i
    double complex COMPLEX_1 # 1 + 0i
    double complex COMPLEX_I # 0 + 1i

### Workspace for Bayer-Helms calculations ###

cdef struct knm_bh_workspace:
    bint matched
    bint aligned
    bint is_mm_changing
    bint is_alignment_changing

    # Beam parameters for coupling from basis q1 -> q2
    complex_t q1
    complex_t q2
    # Factor that applies to a tilt angle which yields
    # the actual misalignment angle gamma
    double beta_factor
    # Misalignment angle between the two systems
    double gamma
    double nr
    double lambda0

    # The Gouy phases of q1 and q2, respectively. These are
    # used in the reverse Gouy calculations.
    double gouy1
    double gouy2

    # The following are mismatch parameters and dependencies
    complex_t K_conj
    complex_t k1
    double K0
    complex_t F
    complex_t FS

    # These contain the misaligment parameters
    complex_t X
    complex_t XS
    double ex # equal to exp(-Re{0.5 * X * XS})
    # pre-calculated cos and sin phases for application
    # of imag part of E^(x) equation
    double ex_cos_phase
    double ex_sin_phase

    # Size of the below caches, equal to 1 + maxtem where
    # maxtem is highest order mode included
    int pow_cache_size
    # Integer power caches for each of {F, FS, X, XS}
    complex_t* F_pow_cache
    complex_t* FS_pow_cache
    complex_t* X_pow_cache
    complex_t* XS_pow_cache
    # And cache for (1 + K0)^{n/2 + 1/4}
    double* K0_plus1_pow_cache

cdef void knm_bh_ws_init(
    knm_bh_workspace* kws,
    complex_t q1,
    complex_t q2,
    double beta,
    double beta_factor,
    double nr,
    double lambda0,
    int maxtem,
    bint is_mismatch_changing=?,
    bint is_alignment_changing=?
) nogil
cdef void knm_bh_ws_free(knm_bh_workspace* kws) nogil

cdef bint knm_bh_ws_is_changing(const knm_bh_workspace* kws) nogil
cdef void knm_bh_ws_recompute(knm_bh_workspace* kws, complex_t q1, complex_t q2, double beta) nogil
cdef void knm_bh_ws_recompute_mismatch(knm_bh_workspace* kws, complex_t q1, complex_t q2) nogil
cdef void knm_bh_ws_recompute_misalignment(knm_bh_workspace* kws, double beta) nogil


### Below are the functions to be used in the compute_scattering_matrices ###
###  function of KnmConnectorWorkspace, each is optimised to the extreme  ###

cdef void fast_compute_refl_knm_matrix_bh__k00_zeroed(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
)

cdef void fast_compute_trns_knm_matrix_bh__k00_zeroed(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
)

cdef void fast_compute_refl_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
)

cdef void fast_compute_trns_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int* homs,
    complex_t* out,
    Py_ssize_t Nhoms,
    int nthreads,
)

### Function for computing BH matrix with no modifications. ###
### The modification functions are defined separately here. ###

cdef void compute_knm_matrix_bh(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const int[:, ::1] homs,
    complex_t[:, ::1] out,
    bint parallel=?
)

cdef void zero_tem00_phase(
    const complex_t[:, ::1] knm_mat,
    complex_t[:, ::1] out
) nogil

cdef void flip_odd_horizontal(
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) nogil

cdef void reverse_gouy_phases(
    const knm_bh_workspace* kws_x, const knm_bh_workspace* kws_y,
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) nogil



from finesse.simulations.basematrix cimport MatrixSystemSolver, CarrierSignalMatrixSimulation

import numpy as np
cimport numpy as np

ctypedef (double*, double*) ptr_tuple_2

cdef class OpticalBandpassValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_2 ptr = (&self.fc, &self.bandwidth)
        cdef tuple params = ("fc", "bandwidth")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class OpticalBandpassConnections:
    """Contains C accessible references to submatrices for
    optical connections for this element.
    """
    def __cinit__(self, object mirror, MatrixSystemSolver mtx):
        # Only 1D arrays of submatrices as no frequency coupling happening
        cdef int Nf = mtx.optical_frequencies.size
        self.P1i_P2o = SubCCSView1DArray(Nf)
        self.P2i_P1o = SubCCSView1DArray(Nf)
        self.ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class OpticalBandpassWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, bint refill):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            OpticalBandpassConnections(owner, sim.carrier),
            OpticalBandpassConnections(owner, sim.signal) if sim.signal else None,
            OpticalBandpassValues()
        )
        # Casting python objects to known types for faster access
        self.cvalues = self.values
        self.carrier_opt_conns = self.carrier.connections
        if sim.signal:
            self.signal_opt_conns = self.signal.connections
        else:
            self.signal_opt_conns = None


cdef void fill_optical_matrix(
        OpticalBandpassWorkspace ws,
        MatrixSystemSolver matrix,
        optical_bandpass_connections *connections
    ):
    cdef double B = 2*np.pi*ws.cvalues.bandwidth
    cdef complex_t H
    cdef frequency_info_t *frequencies = matrix.optical_frequencies.frequency_info
    cdef Py_ssize_t i

    for i in range(matrix.optical_frequencies.size):
        H = 1 / (1 + 1j*(2*np.pi*abs(frequencies[i].f-ws.cvalues.fc))/B)
        if connections.P1i_P2o:
            (<SubCCSView>connections.P1i_P2o[frequencies[i].index]).fill_negative_za_zm_2(
                H, ws.K12.mtx.ptr, ws.K12.mtx.stride1, ws.K12.mtx.stride2
            )

        if connections.P2i_P1o:
            (<SubCCSView>connections.P2i_P1o[frequencies[i].index]).fill_negative_za_zm_2(
                H, ws.K21.mtx.ptr, ws.K21.mtx.stride1, ws.K21.mtx.stride2
            )


optical_bandpass_carrier_fill = FillFuncWrapper.make_from_ptr(c_fill_carrier)
cdef object c_fill_carrier(ConnectorWorkspace ws):
    fill_optical_matrix(
        ws,
        ws.sim.carrier,
        &(<OpticalBandpassWorkspace>ws).carrier_opt_conns.ptrs
    )

optical_bandpass_signal_fill = FillFuncWrapper.make_from_ptr(c_fill_signal)
cdef object c_fill_signal(ConnectorWorkspace ws):
    fill_optical_matrix(
        ws,
        ws.sim.signal,
        &(<OpticalBandpassWorkspace>ws).signal_opt_conns.ptrs
    )


from finesse.cmatrix cimport CCSMatrix
from finesse.cymath cimport complex_t
from finesse.tracing.ctracer cimport TraceForest, TraceTree
from finesse.simulations.base cimport ModelData, NodeBeamParam, NodeInfoEntry, SimConfigData
from finesse.frequency cimport frequency_info_t, FrequencyContainer, Frequency
from cpython.ref cimport PyObject
from finesse.components.workspace cimport ConnectorWorkspace


cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double complex COMPLEX_0


cdef class MatrixSystemWorkspaces:
    cdef readonly:
        list to_initial_fill
        list to_refill
        list to_rhs_refill
        list to_noise_refill
        list to_noise_input_refill
        list noise_detectors
        int num_to_refill
        int num_to_rhs_refill
        int num_to_noise_refill
        int num_to_noise_input_refill
        int num_noise_detectors
    cdef:
        PyObject** ptr_to_refill
        PyObject** ptr_to_rhs_refill
        PyObject** ptr_to_noise_refill
        PyObject** ptr_to_noise_input_refill
        PyObject** ptr_noise_detectors


cdef class MatrixSystemSolver:
    cdef:
        CCSMatrix _M
        readonly dict _noise_matrices
        readonly dict _submatrices
        readonly dict _noise_submatrices
        Py_ssize_t num_nodes
        readonly dict nodes
        readonly dict nodes_idx
        NodeInfoEntry* _c_node_info
        readonly FrequencyContainer optical_frequencies
        readonly dict mechanical_frequencies
        readonly dict electrical_frequencies
        readonly tuple unique_elec_mech_fcnts # Unique frequency containers for mech/elec
        readonly dict noise_sources
        readonly int nhoms
        readonly complex_t[::1] out_view
        readonly bint any_frequencies_changing
        bint is_signal_matrix
        readonly bint forced_refill
        public bint manual_rhs
        readonly MatrixSystemWorkspaces workspaces

    cpdef setup_nodes(self, list nodes)
    cpdef clear_rhs(self)
    cdef initial_fill(self)
    cdef refill(self)
    cdef fill_rhs(self)
    cdef fill_noise_inputs(self)
    cdef refactor(self)
    cdef factor(self)
    cdef solve(self)
    cdef solve_noises(self)
    cdef construct(self)
    cdef initial_run(self)
    cpdef run(self)

    cdef get_node_matrix_params(self, node, Py_ssize_t *Ns, Py_ssize_t *Nf, frequency_info_t** fptr)
    cdef tuple get_node_frequencies(self, node)

    cdef update_frequency_info(self)
    cdef assign_submatrices(self, workspaces)
    cdef assign_noise_submatrices(self, workspaces)

    cdef add_noise_matrix(self, object key)
    cdef add_rhs(self, unicode key)

    cpdef Py_ssize_t findex(self, object node, Py_ssize_t freq)
    cdef Py_ssize_t findex_fast(self, Py_ssize_t node_id, Py_ssize_t freq) nogil

    cpdef Py_ssize_t field(self, object node, Py_ssize_t freq=?, Py_ssize_t hom=?)
    cdef Py_ssize_t field_fast(self, Py_ssize_t node_id, Py_ssize_t freq=?, Py_ssize_t hom=?) nogil
    cdef inline Py_ssize_t field_fast_2(
        self,
        Py_ssize_t node_rhs_idx,
        Py_ssize_t num_hom,
        Py_ssize_t freq,
        Py_ssize_t hom
    ) nogil

    cpdef complex_t get_out(self, object node, Py_ssize_t freq=?, Py_ssize_t hom=?)
    cdef complex_t get_out_fast(self, Py_ssize_t node_id, Py_ssize_t freq=?, Py_ssize_t hom=?) nogil

    cdef void set_source_fast(self, Py_ssize_t node_id, Py_ssize_t freq_idx, Py_ssize_t hom_idx, complex_t value, unsigned rhs_index=?)
    cdef void set_source_fast_2(self, Py_ssize_t rhs_idx, complex_t value)
    cdef void set_source_fast_3(self, Py_ssize_t rhs_idx, complex_t value, unsigned rhs_index)

    cpdef Py_ssize_t node_id(self, object node)
    cpdef get_node_info(self, name)



cdef class CarrierSignalMatrixSimulation:
    cdef:
        readonly MatrixSystemSolver carrier
        readonly MatrixSystemSolver signal
        readonly bint compute_signals

        readonly unicode name
        readonly set changing_parameters
        readonly set tunable_parameters
        readonly object model
        readonly bint is_modal
        readonly bint do_matrix_solving

        readonly MatrixSystemWorkspaces carrier_ws
        readonly MatrixSystemWorkspaces signal_ws
        readonly list detector_workspaces
        readonly list readout_workspaces

        public list workspaces
        public list variable_workspaces
        readonly list gouy_phase_workspaces

        ### Tracing stuff ###
        public dict cavity_workspaces
        NodeBeamParam* trace
        # The TraceForest of geometrically changing branches. This is an
        # empty forest for any simulation in which geometric parameters
        # are not changing.
        readonly TraceForest trace_forest
        # Node couplings which will have potential mode mismatches,
        # determined from trace_forest via tree intersection searching
        readonly tuple mismatched_node_couplings
        # A dict of {<tuple of newly unstable cavities> : <contingent TraceForest>}
        # required for when a scan results in a geometrically changing cavity becoming
        # unstable -> invalidating self.trace_forest temporarily for that data point
        dict contingent_trace_forests
        bint needs_reflag_changing_q # Used when exiting from unstable cavity regions
        bint retrace

        # List of workspaces for components which scatter modes
        list to_scatter_matrix_compute
        readonly ModelData model_data
        object __weakref__

        SimConfigData config_data

    cdef initialise_model_data(self)
    cdef initialise_sim_config_data(self)

    cpdef initialise_workspaces(self)
    cpdef initialise_noise_matrices(self)
    cpdef initialise_noise_sources(self)
    cpdef initialise_noise_selection_vectors(self)
    cdef initialise_trace_forest(self, optical_nodes)
    cdef void update_all_parameter_values(self)

    cdef void update_cavities(self)
    cdef void compute_knm_matrices(self)
    cdef int set_gouy_phases(self) except -1
    cpdef int modal_update(self) except -1

    cpdef run_carrier(self)
    cpdef run_signal(self)

    # Methods to construct the changing TraceForest for the simulation
    cdef void _determine_changing_beam_params(self, TraceForest forest=?, bint set_tree_node_ids=?)
    cdef void _setup_trace_forest(self, TraceForest forest=?, bint set_tree_node_ids=?)
    cdef void _setup_single_trace_tree(self, TraceTree tree, bint set_tree_node_ids=?)

    # Find the newly unstable cavity instances from the changing forest
    cdef tuple _find_new_unstable_cavities(self)
    cdef TraceForest _initialise_contingent_forest(self, tuple unstable_cavities)

    # Perform the beam trace on the changing TraceForest
    cdef void _propagate_trace(self, TraceTree tree, bint symmetric)
    cdef bint trace_beam(self)
